#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                            }....................
import numpy as np
from betse import pathtree
from betse.exceptions import BetseSimConfException, BetseSimPhaseException
from betse.lib.matplotlib import mplcolormap
from betse.lib.yaml.yamlalias import yaml_alias, yaml_enum_alias
from betse.lib.yaml.abc.yamlabc import YamlFileABC
from betse.science.config.confenum import (
    CellLatticeType, GrnUnpicklePhaseType, IonProfileType, SolverType)
from betse.science.config.export.confcsv import SimConfExportCSVs
from betse.science.config.export.visual.confanim import SimConfAnimAll
from betse.science.config.export.visual.confplot import SimConfPlotAll
from betse.science.config.grn.confgrn import SimConfGrnFile
from betse.science.config.model.conftis import (
    SimConfCutListItem, SimConfTissueDefault, SimConfTissueListItem)
from betse.science.phase.phaseenum import SimPhaseKind
from betse.science.tissue.event import tisevevolt
# from betse.util.io.log import logs
from betse.util.path import dirs, pathnames
# from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.types import (
    type_check, IterableTypes, SequenceTypes, StrOrNoneTypes)

# ....................{ CLASSES                            }....................
class Parameters(YamlFileABC):
    '''
    Root YAML-backed in-memory and on-disk simulation configuration,
    encapsulating a low-level container of simulation configuration settings
    both loaded from and saved back to a YAML-formatted configuration file.

    Attributes (Solver)
    ----------
    solver_type : SolverType
        Type of **simulation solver** (i.e., numerical technique iteratively
        computing simulation time steps) with which to solve this simulation.

    Attributes (Path: Export)
    ----------
    init_export_dirname : str
        Absolute pathname of the directory containing results exported from this
        simulation's most recent initialization run, guaranteed to exist.
    init_export_dirname_relative : str
        Relative pathname of the directory containing results exported from this
        simulation's most recent initialization run, relative to the absolute
        path of the directory containing this simulation configuration's YAML
        file.
    sim_export_dirname : str
        Absolute pathname of the directory containing results exported from this
        simulation's most recent simulation run, guaranteed to exist.
    sim_export_dirname_relative : str
        Relative pathname of the directory containing results exported from this
        simulation's most recent simulation run, relative to the absolute path
        of the directory containing this simulation configuration's YAML file.

    Attributes (Path: Pickle: Seed)
    ----------
    seed_pickle_basename : str
        Basename of the pickled file providing this simulation's most recently
        seeded cell cluster, relative to the :attr:`init_pickle_dirname`
        directory.

    Attributes (Path: Pickle: Initialization)
    ----------
    init_pickle_basename : str
        Basename of the pickled file providing this simulation's most recent
        initialization run, relative to the :attr:`init_pickle_dirname`
        directory.
    init_pickle_dirname : str
        Absolute dirname of the directory containing this simulation's most
        recently seeded cell cluster *and* initialization run, guaranteed to
        exist.
    init_pickle_dirname_relative : str
        Relative dirname of the directory containing this simulation's most
        recently seeded cell cluster *and* initialization run, relative to the
        :attr:`conf_dirname` directory.

    Attributes (Path: Pickle: Simulation)
    ----------
    sim_pickle_basename : str
        Basename of the pickled file providing this simulation's most recent
        simulation run within the current :attr:`sim_dirname`.
    sim_pickle_dirname : str
        Absolute dirname of the directory containing this simulation's most
        recent simulation run, guaranteed to exist.
    sim_pickle_dirname_relative : str
        Relative dirname of the directory containing this simulation's most
        recent simulation run, relative to the :attr:`conf_dirname` directory.

    Attributes (Path: Pickle: Gene)
    ----------
    grn_pickle_basename : str
        Basename of the pickled file providing this simulation's most recent
        gene regulatory network (GRN) run, relative to the
        :attr:`grn_pickle_dirname` directory.
    grn_pickle_filename : str
        Absolute filename of the pickled file providing this simulation's most
        recent gene regulatory network (GRN) run.
    grn_pickle_dirname : str
        Absolute dirname of the directory containing this simulation's most
        recent gene regulatory network (GRN) run, guaranteed to exist.
    grn_pickle_dirname_relative : str
        Relative dirname of the directory containing this simulation's most
        recent gene regulatory network (GRN) run, relative to the
        :attr:`conf_dirname` directory.
    grn_unpickle_filename : StrOrNoneTypes
        Absolute filename of the pickled file providing a prior gene regulatory
        network (GRN) run for this simulation if restarting the current such run
        from where this prior run left off *or* ``None`` otherwise (i.e., if
        starting the current such run from scratch).
    grn_unpickle_filename_relative : StrOrNoneTypes
        Relative filename of the pickled file providing a prior gene regulatory
        network (GRN) run for this simulation, relative to the
        :attr:`conf_dirname` directory, if restarting the current such run from
        where this prior run left off *or* ``None`` otherwise (i.e., if starting
        the current such run from scratch).

    Attributes (Space: Cell)
    ----------
    cell_radius : float
        Radius in meters of each cell in this cluster. This should typically be
        within an order of magnitude of the recommended default of ``5.0e-6``.

    Attributes (Space: Cell Cluster)
    ----------
    cell_lattice_disorder : float
        Degree to which cell boundaries spatially deviate from the base cell
        lattice. If 0.0, cells rigidly conform to this lattice; else, cells
        randomly deviate from this lattice. Increasing this increases the
        entropy (i.e., randomness) of these deviations, producing an
        increasingly chaotic arrangement of cells. This should typically reside
        in the range ``[0.0, 0.8]``.
    cell_lattice_type : CellLatticeType
        Type of **base cell lattice** (i.e., uniform grid to which cells are
        situated *before* random lattice disorder is applied).

    Attributes (Space: Environment)
    ----------
    grid_size : int
        Number of square grid spaces (in both the X and Y dimensions) to
        computationally divide this simulation's square environment into.
        Increasing this increases simulation granularity and hence stability at
        a quadratic increase in space and time costs (i.e., ``O(grid_size **
        2)``). This should typically reside in the range ``[10, 60]``.
    is_ecm : bool
        ``True`` only if the extracellular matrix (ECM) simulating
        **extracellular spaces** (i.e., the environment surrounding each cell in
        the cluster) is enabled. Disabling this reduces simulation accuracy at a
        substantial reduction in space and time costs.
    world_len : float
        Length in meters of both the X and Y dimensions of this simulation's
        square environment. This should typically reside in the range ``[80e-6,
        1000e-6]``.

    Attributes (Space: Tissue)
    ----------
    cut_profiles : YamlList
        List of all **non-default cut profiles** (i.e., objects targeting a
        region of the cell cluster to be permanently removed by a corresponding
        simulation event). Ignored if :attr:`is_tissue_profiles` is ``False``.
    is_tissue_profiles : bool
        ``True`` only if **tissue profiles** (i.e., user-defined regions within
        the cell cluster to which specific base membrane diffusion profiles,
        interventions, and individualized dynamics may be applied) are enabled.
        Note that tissue profiles should typically *always* be enabled.
    tissue_default : SimConfTissueDefault
        Default tissue profile applied to all cells *not* already targeted by
        another tissue profile in the :attr:`tissue_profiles` list.
    tissue_profiles : YamlList
        List of all **non-default tissue profiles** (i.e., objects targeting a
        region of the cell cluster to be associated with particular simulation
        constants and parameters). Ignored if :attr:`is_tissue_profiles` is
        ``False``.

    Attributes (Time: Total)
    ----------
    init_time_total : float
        Duration in seconds of the initialization phase *not* modified by
        extraneous settings (e.g., gap junction acceleration factor).
    sim_time_total : float
        Duration in seconds of the simulation phase *not* modified by extraneous
        settings (e.g., gap junction acceleration factor).
    total_time : float
        Duration in seconds of the current simulation phase *not* modified by
        extraneous settings (e.g., gap junction acceleration factor).
    total_time_accelerated : float
        Duration in seconds of the current simulation phase accelerated by the
        gap junction acceleration factor.

    Attributes (Time: Step)
    ----------
    init_time_step : float
        Duration in seconds of each time step (including sampled and unsampled)
        for the initialization phase.
    sim_time_step : float
        Duration in seconds of each time step (including sampled and unsampled)
        for the initialization phase.
    dt : float
        Duration in seconds of each time step (including sampled and unsampled)
        for the current simulation phase.

    Attributes (Time: Sampling)
    ----------
    init_time_sampling : float
        Duration in seconds between each sampled time step (including that
        sampled time step itself) for the initialization phase. Decreasing this
        duration increases the number of time steps for which data is exported
        from this phase at a linear cost in space consumption.
    sim_time_sampling : float
        Duration in seconds between each sampled time step (including that
        sampled time step itself) for the simulation phase. Decreasing this
        duration increases the number of time steps for which data is exported
        from this phase at a linear cost in space consumption.
    t_resample : float
        Number of time steps between each sampled time step, including that
        sampled time step itself. Notably, if the current time step ``t`` is a
        sampled time step:
        * ``t + t_resample`` is the next sampled time step (if any).
        * ``t - t_resample`` is the prior sampled time step (if any).

    Attributes (Gene Regulatory Network)
    ----------
    The following attributes are ignored if :attr:`grn_enabled` is ``False``.

    grn : SimConfGrnFile
        Gene regulatory network (GRN) subconfiguration, encapsulating *all*
        GRN-related settings both loaded from and saved back to the separate
        YAML-formatted configuration file with filename
        :attr:`grn_config_filename`.
    grn_unpickle_phase_type : GrnUnpicklePhaseType
        Type of **unpickle simulation phase** (i.e., previously pickled
        simulation phase to unpickle as the computational basis for the current
        network to be run by the ``betse sim-grn`` subcommand).

    Attributes (Ion)
    ----------
    ion_profile : IonProfileType
        Type of **ion profile** (i.e., predefined set of all extracellular and
        cytosolic ions enabled by this simulation).
    ions_dict : dict
        Dictionary mapping:
        * From each of the following names of a supported core ion:
          * ``Na``, the sodium cation Na+.
          * ``K``, the potassium cation K+.
          * ``Ca``, the calcium cation Ca2+.
          * ``Cl``, the chloride anion Cl-.
          * ``H``, the hydrogen cation H+.
          * ``M``, the M anion bicarbonate HCO3-.
          * ``P``, the anionic protein P-.
        * To either:
          * 0, if that ion is disabled by the ion profile in the current
            simulation configuration.
          * 1, if that ion is enabled by that ion profile.

    Attributes (Ion: Initial)
    ----------
    conc_env_ca : float
        Initial extracellular concentration of the Ca2+ (calcium) cation.
    cCl_env : float
        Initial extracellular concentration of the Cl- (chloride) anion.
    conc_env_k : float
        Initial extracellular concentration of the K+ (potassium) cation.
    conc_env_m : float
        Initial extracellular concentration of the M- (unidentified) anion,
        synthetically manufactured to enforce charge-balancing over all
        environmental spaces for both simulation stability and correctness.
    conc_env_na : float
        Initial extracellular concentration of the Na+ (sodium) cation.
    conc_env_p : float
        Initial extracellular concentration of the P- (protein) anion.

    Attributes (Ion: Initial: Custom)
    ----------

    Attributes (General: Scalars)
    ----------
    cell_polarizability : NumericSimpleTypes
        Constant defining the rate of cell polarizability change in electric
        fields, typically ranging ``[0.0, 1.0e-3]``.

    Attributes (Exports)
    ----------
    anim : SimConfAnimAll
        Subconfiguration configuring exported animations.
    csv : SimConfExportCSVs
        Subconfiguration configuring exported comma-separated value (CSV) files.
    plot : SimConfPlotAll
        Subconfiguration configuring exported plots.
    plot_cell : int
        0-based index of the cell to isolate all single-cell time plots to.
        Defaults to 0, the index assigned to the first cell guaranteed to exist.
        Note that cell indices are seed-specific and may be visualized via the
        :attr:`enumerate_cells` boolean.
    '''

    # ..................{ ALIASES                            }..................
    # To sanitize computation throughout the codebase, *ALL* real numbers are
    # required to be floating point rather the more general "NumericSimpleTypes" type
    # (i.e., either floating point or integer). Due to magic internal to the
    # yaml_alias() data descriptor, integer values are both silently and safely
    # cast to floating point values.

    # ..................{ ALIASES ~ solver                   }..................
    solver_type = yaml_enum_alias("['solver options']['type']", SolverType)

    # ..................{ ALIASES ~ path : seed              }..................
    seed_pickle_basename = yaml_alias("['init file saving']['worldfile']", str)

    # ..................{ ALIASES ~ path : init              }..................
    init_pickle_basename = yaml_alias("['init file saving']['file']", str)
    init_pickle_dirname_relative = yaml_alias(
        "['init file saving']['directory']", str)
    init_export_dirname_relative = yaml_alias(
        "['results file saving']['init directory']", str)

    # ..................{ ALIASES ~ path : sim               }..................
    sim_pickle_basename = yaml_alias("['sim file saving']['file']", str)
    sim_pickle_dirname_relative = yaml_alias(
        "['sim file saving']['directory']", str)
    sim_export_dirname_relative  = yaml_alias(
        "['results file saving']['sim directory']", str)

    # ..................{ ALIASES ~ path : grn               }..................
    grn_pickle_basename = yaml_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['save to file']", str)
    grn_pickle_dirname_relative = yaml_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['save to directory']", str)
    grn_unpickle_filename_relative = yaml_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['load from']", StrOrNoneTypes)

    # ..................{ ALIASES ~ space : cell             }..................
    cell_radius = yaml_alias("['world options']['cell radius']", float)

    # ..................{ ALIASES ~ space : cell cluster     }..................
    cell_lattice_disorder = yaml_alias(
        "['world options']['lattice disorder']", float)
    cell_lattice_type = yaml_enum_alias(
        "['world options']['lattice type']", CellLatticeType)

    # ..................{ ALIASES ~ space : env              }..................
    grid_size = yaml_alias("['general options']['comp grid size']", int)
    is_ecm = yaml_alias(
        "['general options']['simulate extracellular spaces']", bool)
    world_len = yaml_alias("['world options']['world size']", float)

    # ..................{ ALIASES ~ space : tissue           }..................
    #FIXME: Does this boolean actually serve a demonstrable purpose? I might be
    #offbase here, but don't we always want tissue profiles? Is there actually a
    #useful use case for even disabling all tissue profiles?
    is_tissue_profiles = yaml_alias(
        "['tissue profile definition']['profiles enabled']", bool)

    # ..................{ ALIASES ~ time : init              }..................
    init_time_total = yaml_alias("['init time settings']['total time']", float)
    init_time_step = yaml_alias("['init time settings']['time step']", float)
    init_time_sampling = yaml_alias(
        "['init time settings']['sampling rate']", float)

    # ..................{ ALIASES ~ time : sim               }..................
    sim_time_total  = yaml_alias("['sim time settings']['total time']", float)
    sim_time_step  = yaml_alias("['sim time settings']['time step']", float)
    sim_time_sampling = yaml_alias(
        "['sim time settings']['sampling rate']", float)

    # ..................{ ALIASES ~ grn                      }..................
    grn_unpickle_phase_type = yaml_enum_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['run network on']",
        GrnUnpicklePhaseType)

    # ..................{ ALIASES ~ ion                      }..................
    #FIXME: Consider shifting all ion-centric functionality into a dedicated
    #"ion" instance variable, instantiated to be an instance of a newly defined
    #"SimConfIonProfile" class. Ion handling currently appears to consume in
    #upwards of a third of this entire submodule.
    ion_profile = yaml_enum_alias(
        "['general options']['ion profile']", IonProfileType)

    # ..................{ ALIASES ~ ion ~ custom             }..................
    #FIXME: Actually use this.
    ion_profile_custom_conc_env_na = yaml_alias(
        "['general options']['customized ion profile']"
        "['extracellular Na+ concentration']", float)

    # ..................{ ALIASES ~ scalar                   }..................
    cell_polarizability = yaml_alias(
        "['internal parameters']['cell polarizability']", float)

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        #FIXME: The following should also be performed by the unload() method,
        #suggesting that this logic should be shifted there and the unload()
        #method then explicitly called -- perhaps by our superclass?

        # Nullify all instance variables for safety.
        self.grn_pickle_dirname = None
        self.grn_pickle_filename = None
        self.grn_unpickle_filename = None
        self.init_export_dirname = None
        self.init_pickle_dirname = None
        self.sim_export_dirname = None
        self.sim_pickle_dirname = None

        # Classify unloaded tissue and cut profiles.
        self.cut_profiles = SimConfCutListItem.make_list()
        self.tissue_default = SimConfTissueDefault()
        self.tissue_profiles = SimConfTissueListItem.make_list()

        # Classify unloaded export subconfigurations.
        self.anim = SimConfAnimAll()
        self.csv = SimConfExportCSVs()
        self.plot = SimConfPlotAll()

        # Classify unloaded GRN subconfigurations.
        self.grn = SimConfGrnFile()

    # ..................{ LOADERS                            }..................
    #FIXME: Convert all or most of the variables parsed by this method into
    #aliases of the above form. Brainy rainbows!
    @type_check
    def load(self, *args, **kwargs) -> YamlFileABC:

        # Avoid circular import dependencies.
        from betse.science.compat import compatconf

        # Defer to the superclass implementation.
        super().load(*args, **kwargs)

        # Preserve backward compatibility with prior configuration formats
        # *BEFORE* other initialization, which expects the passed YAML file to
        # conform to the current configuration format.
        compatconf.upgrade_sim_conf(self)

        # Initialize paths specified by this configuration.
        self.reload_paths()

        # Load all currently unloaded tissue subconfigurations.
        self.cut_profiles.load(
            conf=self._conf['tissue profile definition']['cut profiles'])
        self.tissue_default.load(
            conf=self._conf['tissue profile definition']['tissue']['default'])
        self.tissue_profiles.load(
            conf=self._conf['tissue profile definition']['tissue']['profiles'])

        # Load all currently unloaded export subconfigurations.
        self.anim.load(conf=self._conf)
        self.csv.load(conf=self._conf)
        self.plot.load(conf=self._conf)

        #---------------------------------------------------------------------------------------------------------------
        # GENERAL OPTIONS
        #---------------------------------------------------------------------------------------------------------------

        self.autoInit = self._conf['automatically run initialization']
        self.plot_grid_size = 50

        #---------------------------------------------------------------------------------------------------------------
        # WORLD OPTIONS
        #---------------------------------------------------------------------------------------------------------------

        # Geometric constants and factors
        self.wsx = self.world_len  # the x-dimension of the world space [m]
        self.wsy = self.world_len  # the y-dimension of the world space [m]
        self.cell_height = float(self._conf['world options']['cell height'])  # the height of a cell in the z-direction
        self.cell_space = float(self._conf['world options']['cell spacing'])  # the true cell-cell spacing

        volmult = float(self._conf['internal parameters']['environment volume multiplier'])

        self.vol_env = volmult*self.wsx*self.wsy*self.cell_height  # environmental volume for "no ECM" simulation

        # Parameters for Lloyd's Voronoi mesh optimization settings during seed:
        mesh_refine = self._conf['world options'].get('mesh refinement', None)

        if mesh_refine is not None:
            self.refine_mesh = mesh_refine['refine mesh']
            self.maximum_voronoi_steps = int(mesh_refine['maximum steps'])
            self.voronoi_convergence = float(mesh_refine['convergence threshold'])

        else:
            self.refine_mesh = False
            self.maximum_voronoi_steps = 20
            self.voronoi_convergence = 1.0e-9





        #---------------------------------------------------------------------------------------------------------------
        # TARGETED INTERVENTIONS
        #---------------------------------------------------------------------------------------------------------------

        #FIXME: This dictionary appears to be superfluous. It's never iterated
        #over or treated like a dictionary. Its keys are only ever looked up. In
        #short, this dictionary shouldn't exist. For efficiency and sanity, all
        #keys of this dictionary should be refactored into simple instance
        #variables of this class (e.g., from "p.scheduled_options['Na_mem']" to
        #simply "p.event_Na_mem").
        #FIXME: After doing so, these instance variables should be refactored
        #from unreadable lists to objects with human-readable attribute names
        #(e.g., from "p.event_Na_mem[0]" to "p.event_Na_mem.start_time").
        #
        #The longest journey to the future begins tomorrow.

        # initialize dictionary keeping track of targeted scheduled options for the sim:
        self.scheduled_options = {}

        bool_Namem = bool(self._conf['change Na mem']['event happens'])
        bool_Kmem = bool(self._conf['change K mem']['event happens'])
        bool_Clmem = bool(self._conf['change Cl mem']['event happens'])
        bool_Camem = bool(self._conf['change Ca mem']['event happens'])
        bool_ecmj = bool(self._conf['break ecm junctions']['event happens'])
        bool_pressure = bool(self._conf['apply pressure']['event happens'])

        if bool_Namem is False:
            self.scheduled_options['Na_mem'] = 0
        elif bool_Namem is True:
            on_Namem = float(self._conf['change Na mem']['change start'])
            off_Namem = float(self._conf['change Na mem']['change finish'])
            rate_Namem = float(self._conf['change Na mem']['change rate'])
            multi_Namem = float(self._conf['change Na mem']['multiplier'])
            apply_Namem = self._conf['change Na mem']['apply to']
            function = self._conf['change Na mem']['modulator function']
            Namem = [on_Namem, off_Namem, rate_Namem, multi_Namem, apply_Namem,function]
            self.scheduled_options['Na_mem'] = Namem

        if bool_Kmem is False:
            self.scheduled_options['K_mem'] = 0
        elif bool_Kmem is True:
            on_Kmem = float(self._conf['change K mem']['change start'])
            off_Kmem = float(self._conf['change K mem']['change finish'])
            rate_Kmem = float(self._conf['change K mem']['change rate'])
            multi_Kmem = float(self._conf['change K mem']['multiplier'])
            apply_Kmem = self._conf['change K mem']['apply to']
            function = self._conf['change K mem']['modulator function']
            Kmem = [on_Kmem, off_Kmem, rate_Kmem, multi_Kmem, apply_Kmem,function]
            self.scheduled_options['K_mem'] = Kmem

        if bool_Clmem is False:
            self.scheduled_options['Cl_mem'] = 0
        elif bool_Clmem is True:
            on_Clmem = float(self._conf['change Cl mem']['change start'])
            off_Clmem = float(self._conf['change Cl mem']['change finish'])
            rate_Clmem = float(self._conf['change Cl mem']['change rate'])
            multi_Clmem = float(self._conf['change Cl mem']['multiplier'])
            apply_Clmem = self._conf['change Cl mem']['apply to']
            function = self._conf['change Cl mem']['modulator function']
            Clmem = [on_Clmem, off_Clmem, rate_Clmem, multi_Clmem, apply_Clmem, function]
            self.scheduled_options['Cl_mem'] = Clmem

        if bool_Camem is False:
            self.scheduled_options['Ca_mem'] = 0
        elif bool_Camem is True:
            on_Camem = float(self._conf['change Ca mem']['change start'])
            off_Camem = float(self._conf['change Ca mem']['change finish'])
            rate_Camem = float(self._conf['change Ca mem']['change rate'])
            multi_Camem = float(self._conf['change Ca mem']['multiplier'])
            apply_Camem = self._conf['change Ca mem']['apply to']
            function = self._conf['change Ca mem']['modulator function']
            Camem = [on_Camem, off_Camem, rate_Camem, multi_Camem, apply_Camem,function]
            self.scheduled_options['Ca_mem'] = Camem

        if bool_pressure:
            on_p = float(self._conf['apply pressure']['change start'])
            off_p = float(self._conf['apply pressure']['change finish'])
            rate_p = float(self._conf['apply pressure']['change rate'])
            multi_p = float(self._conf['apply pressure']['multiplier'])
            apply_p = self._conf['apply pressure']['apply to']
            function = self._conf['apply pressure']['modulator function']
            pressure_ops = [on_p, off_p, rate_p, multi_p, apply_p,function]

            self.scheduled_options['pressure'] = pressure_ops
        else:
            self.scheduled_options['pressure'] = 0

        #FIXME: Rename this dictionary key from "extV" to "external voltage".
        # Parameterize the voltage event if enabled.
        self.scheduled_options['extV'] = tisevevolt.make(p=self)

        if bool_ecmj is False:
            self.scheduled_options['ecmJ'] = 0
        elif bool_ecmj is True:
            on_ecmj = float(self._conf['break ecm junctions']['change start'])
            off_ecmj = float(self._conf['break ecm junctions']['change finish'])
            rate_ecmj = float(self._conf['break ecm junctions']['change rate'])
            apply_ecmj = self._conf['break ecm junctions']['apply to']
            mult_ecmj = 1.0 - float(self._conf['break ecm junctions'].get('multiplier', 0.0))
            ecmj = [on_ecmj, off_ecmj, rate_ecmj, apply_ecmj, mult_ecmj]

            self.scheduled_options['ecmJ'] = ecmj

        # Parameterize the cutting event if enabled.
        self.break_TJ = self._conf['cutting event'].get('break TJ', True)
        self.wound_TJ = float(self._conf['cutting event'].get('wound TJ', 0.1))
        self.cut_time = float(self._conf['cutting event'].get('cut time', 0.0))


        #---------------------------------------------------------------------------------------------------------------
        # GLOBAL INTERVENTIONS
        #---------------------------------------------------------------------------------------------------------------

        # initialize dictionary keeping track of global scheduled options for the sim:
        self.global_options = {}

        bool_Naenv = bool(self._conf['change Na env']['event happens'])
        bool_Kenv = bool(self._conf['change K env']['event happens'])
        bool_Clenv = bool(self._conf['change Cl env']['event happens'])
        bool_gjblock = bool(self._conf['block gap junctions']['event happens'])
        bool_temp =  bool(self._conf['change temperature']['event happens'])
        bool_NaKblock = bool(self._conf['block NaKATP pump']['event happens'])

        if bool_Kenv is False:
            self.global_options['K_env'] = 0
        elif bool_Kenv is True:
            on_Kenv = float(self._conf['change K env']['change start'])
            off_Kenv = float(self._conf['change K env']['change finish'])
            rate_Kenv = float(self._conf['change K env']['change rate'])
            multi_Kenv = float(self._conf['change K env']['multiplier'])
            kenv = [on_Kenv, off_Kenv, rate_Kenv, multi_Kenv]
            self.global_options['K_env'] = kenv

        if bool_Clenv is False:
            self.global_options['Cl_env'] = 0
        elif bool_Clenv is True:
            on_Clenv = float(self._conf['change Cl env']['change start'])
            off_Clenv = float(self._conf['change Cl env']['change finish'])
            rate_Clenv = float(self._conf['change Cl env']['change rate'])
            multi_Clenv = float(self._conf['change Cl env']['multiplier'])
            Clenv = [on_Clenv, off_Clenv, rate_Clenv, multi_Clenv]
            self.global_options['Cl_env'] = Clenv

        if bool_Naenv is False:
            self.global_options['Na_env'] = 0
        elif bool_Naenv is True:
            on_Naenv = float(self._conf['change Na env']['change start'])
            off_Naenv = float(self._conf['change Na env']['change finish'])
            rate_Naenv = float(self._conf['change Na env']['change rate'])
            multi_Naenv = float(self._conf['change Na env']['multiplier'])
            Naenv = [on_Naenv, off_Naenv, rate_Naenv, multi_Naenv]
            self.global_options['Na_env'] = Naenv

        if bool_gjblock is False:
            self.global_options['gj_block'] = 0
        elif bool_gjblock is True:
            on_gj = float(self._conf['block gap junctions']['change start'])
            off_gj = float(self._conf['block gap junctions']['change finish'])
            rate_gj = float(self._conf['block gap junctions']['change rate'])
            fraction_gj = float(self._conf['block gap junctions']['random fraction'])
            gjb = [on_gj,off_gj,rate_gj,fraction_gj]
            self.global_options['gj_block'] = gjb

        if bool_temp is False:
            self.global_options['T_change'] = 0
        elif bool_temp is True:
            on_T = float(self._conf['change temperature']['change start'])
            off_T = float(self._conf['change temperature']['change finish'])
            rate_T = float(self._conf['change temperature']['change rate'])
            multi_T = float(self._conf['change temperature']['multiplier'])
            temper = [on_T, off_T, rate_T, multi_T]
            self.global_options['T_change'] = temper

        if bool_NaKblock is False:
            self.global_options['NaKATP_block'] = 0
        elif bool_NaKblock is True:
            on_nak = float(self._conf['block NaKATP pump']['change start'])
            off_nak = float(self._conf['block NaKATP pump']['change finish'])
            rate_nak = float(self._conf['block NaKATP pump']['change rate'])
            nak = [on_nak,off_nak,rate_nak]
            self.global_options['NaKATP_block'] = nak

        # Calcium TissueHandler: Calcium Induced Calcium Release (CICR).....................................................

        # include full calcium dynamics in the situation (i.e. endoplasmic reticulum, etc)?
        self.Ca_dyn = False

        #--------------------------------------------------------------------------------------------------------------
        #  CUSTOM GENERAL NETWORK (defined in main config file)
        #--------------------------------------------------------------------------------------------------------------

        self.network_config = self._conf.get('general network', None)

        if self.network_config is not None:
            self.molecules_enabled = self.network_config['implement network']
            self.mol_mit_enabled = self.network_config['enable mitochondria']

        else:
            self.mol_mit_enabled = False

        #---------------------------------------------------------------------------------------------------------------
        #  GENE REGULATORY NETWORKS
        #---------------------------------------------------------------------------------------------------------------

        self.grn_enabled = self._conf['gene regulatory network settings']['gene regulatory network simulated']

        self.grn_config_filename = self._conf['gene regulatory network settings']['gene regulatory network config']

        # If a GRN is enabled, load this GRN from this file.
        if self.grn_enabled:
            self.grn.load(conf_filename=self.grn_config_filename)

        simgrndic = (
            self._conf['gene regulatory network settings']['sim-grn settings'])

        self.grn_dt = float(simgrndic.get('time step', 1.0e-2))
        self.grn_total_time = float(simgrndic.get('total time', 10.0))
        self.grn_tsample = float(simgrndic.get('sampling rate', 1.0))
        self.grn_runmodesim = simgrndic.get('run as sim', False)

        #--------------------------------------------------------------------------------------------------------------
        # VARIABLE SETTINGS
        #--------------------------------------------------------------------------------------------------------------

        self.T = float(self._conf['variable settings']['temperature'])  # system temperature

        # use the GHK equation to calculate alt Vmem from params?
        self.GHK_calc = self._conf['variable settings']['use Goldman calculator']

        # electroosmotic fluid flow-----------------------------------------------------
        self.fluid_flow = self._conf['variable settings']['fluid flow']['include fluid flow']
        self.mu_water = float(self._conf['variable settings']['fluid flow']['water viscocity']) # visc water [Pa.s]

        # electrodiffusive movement pumps and channels -----------------------------------
        self.sim_eosmosis = False
        # self.D_membrane = float(self._conf['variable settings']['channel electroosmosis']['membrane mobility'])
        # self.z_channel = float(self._conf['variable settings']['channel electroosmosis']['channel charge'])
        # self.z_pump = float(self._conf['variable settings']['channel electroosmosis']['pump charge'])

        # mechanical deformation ----------------------------------------------------------
        self.deformation = self._conf['variable settings']['deformation']['turn on']

        self.galvanotropism = float(self._conf['variable settings']['deformation']['galvanotropism'])
        self.td_deform = False # this has been disabled due to ongoing technical difficulties
        self.fixed_cluster_bound = self._conf['variable settings']['deformation']['fixed cluster boundary']
        self.youngMod = float(self._conf['variable settings']['deformation']['young modulus'])
        self.mu_tissue = float(self._conf['variable settings']['deformation']['viscous damping'])

        # osmotic and electrostatic pressures --------------------------------
        self.deform_osmo = self._conf['variable settings']['pressures']['include osmotic pressure']
        self.aquaporins = float(self._conf['variable settings']['pressures']['membrane water conductivity'])

        # calculate lame's parameters from young mod and the poisson ratio:
        self.poi = 0.49 # Poisson's ratio for the biological medium
        self.lame_mu = self.youngMod/(2*(1+self.poi))
        self.lame_lamb = (self.youngMod*self.poi)/((1+self.poi)*(1-2*self.poi))
        self.mu_membrane = 1.0 # membrane viscocity
        self.zeta = -70e-3  # zeta potential of cell membrane [V]

        # Gap junction parameters ------------------
        self.gj_surface = float(self._conf['variable settings']['gap junctions']['gap junction surface area'])
        self.gj_flux_sensitive = False
        self.gj_vthresh = float(self._conf['variable settings']['gap junctions']['gj voltage threshold'])
        self.gj_vgrad  = float(self._conf['variable settings']['gap junctions']['gj voltage window'])
        self.gj_min = float(self._conf['variable settings']['gap junctions']['gj minimum'])
        self.gj_respond_flow = False # (feature currently unsupported)
        self.v_sensitive_gj = self._conf['variable settings']['gap junctions']['voltage sensitive gj']

        # Microtubule properties........................................................................................

        mtb = self.config['variable settings'].get('microtubules', {})

        # if mtb is not None:

        self.use_microtubules = mtb.get('use microtubules', True)

        self.length_charge = float(mtb.get('charge per micrometer', -360.0))
        self.mt_radius = float(mtb.get('radius', 15.0e-9))
        self.mt_length = float(mtb.get('length', 5.0e-6))
        self.tubulin_dipole = float(mtb.get('tubulin unit dipole', 1750))
        self.tubulin_polar = float(mtb.get('tubulin polarizability', 50.0))

        self.tethered_tubule = (mtb.get('tethered tubule', True))

        self.cytoplasm_viscocity = 1.0e-2

        self.D_mtube = float(mtb.get('microtubule diffusion', 1.0))

        self.dilate_mtube_dt = float(mtb.get('time dilation factor', 1.0))

        self.mtube_init_x = mtb.get('initial x-coorinate', None)   # bitmap forcing alignment as initial state
        self.mtube_init_y = mtb.get('initial y-coorinate', None)  # bitmap forcing alignment as initial state
        self.mtube_init_rotangle = float(mtb.get('rotate initialization axis', 0.0))  # angle (in degrees) rotating axis

        if self.mtube_init_x == 'None':
            self.mtube_init_x = None

        if self.mtube_init_y == 'None':
            self.mtube_init_y = None

        if self.mtube_init_rotangle == 'None':
            self.mtube_init_rotangle = 0.0

        self.mt_space_density = mtb.get('microtubule density', None)

        if self.mt_space_density == 'None':

            self.mt_space_density = None

        self.microtubules_orient_parallel = mtb.get('microtubules orient parallel', True)


        #...............................................................................................................

        # Environmental features and tight junctions ---------------------------------------------------

        #FIXME: This doesn't appear to be used anywhere, at the moment. Is this
        #safely removable now, or did we have Big Plans for this at some point?
        self.env_type = True # for now, can't handle air boundaries

        #FIXME: Should this actually be configurable? If not, no worries! -.-
        self.cluster_open = True

        self.D_tj = float(self._conf['variable settings']['tight junction scaling'])
        self.D_adh = float(self._conf['variable settings']['adherens junction scaling'])
        # tight junction relative ion movement properties:
        self.Dtj_rel = {}  # use a dictionary to hold the tj values:

        self.Dtj_rel['Na']=float(self._conf['variable settings']['tight junction relative diffusion']['Na'])
        self.Dtj_rel['K']=float(self._conf['variable settings']['tight junction relative diffusion']['K'])
        self.Dtj_rel['Cl']=float(self._conf['variable settings']['tight junction relative diffusion']['Cl'])
        self.Dtj_rel['Ca']=float(self._conf['variable settings']['tight junction relative diffusion']['Ca'])
        self.Dtj_rel['M']=float(self._conf['variable settings']['tight junction relative diffusion']['M'])
        self.Dtj_rel['P']=float(self._conf['variable settings']['tight junction relative diffusion']['P'])

        # environmental (global) boundary concentrations:
        self.cbnd = self._conf['variable settings']['env boundary concentrations']

        # include noise in the simulation?
        self.channel_noise_level = float(self._conf['variable settings']['noise']['static noise level'])

        self.dynamic_noise = self._conf['variable settings']['noise']['dynamic noise']
        self.dynamic_noise_level = float(self._conf['variable settings']['noise']['dynamic noise level'])

        # Modulator functions ------------------------------------------------------------------------------------------
        self.gradient_x_properties = {}
        self.gradient_y_properties = {}
        self.gradient_r_properties = {}

        self.periodic_properties = {}
        self.f_scan_properties = {}

        self.gradient_x_properties['slope'] =float(self._conf['modulator function properties']['gradient_x']['slope'])
        self.gradient_x_properties['x-offset'] =float(self._conf['modulator function properties']['gradient_x'].get('x-offset', 0.0))
        self.gradient_x_properties['z-offset'] =float(self._conf['modulator function properties']['gradient_x'].get('z-offset', 0.0))
        self.gradient_x_properties['exponent'] = float(
                                        self._conf['modulator function properties']['gradient_x'].get('exponent', 1))

        self.gradient_y_properties['slope'] =float(self._conf['modulator function properties']['gradient_y']['slope'])
        self.gradient_y_properties['x-offset'] = float(self._conf['modulator function properties']['gradient_y'].get('x-offset', 0.0))
        self.gradient_y_properties['z-offset'] = float(self._conf['modulator function properties']['gradient_y'].get('z-offset', 0.0))
        self.gradient_y_properties['exponent'] = float(
                                    self._conf['modulator function properties']['gradient_y'].get('exponent', 1))

        self.gradient_r_properties['slope'] = float(self._conf['modulator function properties']['gradient_r']['slope'])
        self.gradient_r_properties['x-offset'] = float(self._conf['modulator function properties']['gradient_r'].get('x-offset', 0.0))
        self.gradient_r_properties['z-offset'] = float(self._conf['modulator function properties']['gradient_r'].get('z-offset', 0.0))
        self.gradient_r_properties['exponent'] = float(self._conf['modulator function properties']['gradient_r'].get('exponent', 1))

        self.periodic_properties['frequency'] = float(self._conf['modulator function properties']['periodic']['frequency'])
        self.periodic_properties['phase'] = float(self._conf['modulator function properties']['periodic']['phase'])

        self.f_scan_properties['f start'] = \
                                float(self._conf['modulator function properties']['f_sweep']['start frequency'])

        self.f_scan_properties['f stop'] = \
                                float(self._conf['modulator function properties']['f_sweep']['end frequency'])

        #initialize the f vect field to None as it's set depending on the sim timestep:

        self.f_scan_properties['f slope'] = None

        # filename to read for bitmap loading gradient definition:
        chk = self._conf['modulator function properties'].get('gradient_bitmap', None)

        if chk is not None:
            self.grad_bm_fn = self._conf['modulator function properties']['gradient_bitmap']['file']
            self.grad_bm_offset = self._conf['modulator function properties']['gradient_bitmap'].get('z-offset', 0.0)

        else:
            self.grad_bm_fn = None
            self.grad_bm_offset = None

        # ................{ EXPORTS                            }................
        ro = self._conf['results options']

        # ................{ EXPORTS ~ plot                     }................
        #FIXME: Replace all instances of "p.turn_all_plots_off" in the codebase
        #by "not p.plot.is_after_sim_show" and remove this attribute entirely.
        self.turn_all_plots_off = not self.plot.is_after_sim_show

        #FIXME: Replace all instances of "p.autosave" in the codebase
        #by "p.plot.is_after_sim_save" and remove this attribute entirely.
        self.autosave = self.plot.is_after_sim_save  # autosave all still images to a results directory

        self.plot_cutlines = ro['plot cutlines']

        # Colormaps.
        self.default_cm    = mplcolormap.get_colormap(ro['default colormap'])
        self.background_cm = mplcolormap.get_colormap(ro['background colormap'])

        # Colormap for plotting gj currents on top of default colormap.
        self.gj_cm = mplcolormap.get_colormap(ro['gj colormap'])

        # new options for plotting reaction network graphs:
        self.plot_network = ro['plot networks']
        self.network_cm = mplcolormap.get_colormap(ro['network colormap'])

        # Colors.
        self.vcolor = ro['vector and stream color']  # color of vector and streamlines

        # True if numbering cells in plots and animations.
        self.enumerate_cells = ro['enumerate cells']

        # FIXME, let this be a list!
        self.plot_cell = ro['plot cell index']             # State the cell index to use for single-cell time plots

        self.plot_networks_single_cell = ro['plot networks single cell']
        self.showCells = ro['show cells']     # True = polygon patch plots, False = trimesh
        self.stream_density = ro['streamline density']
        self.IecmPlot = ro['plot total current']    # True = plot extracellular currents, false plot gj
        self.plotMask = ro['plot masked geometry']

        #FIXME: Remove all of the following after globally removing "plot seed".
        # Plot seed options.
        self.plot_cluster_mask = ro.get('plot cluster mask', True)

        #--------------------------------------------------------------------------------------------------------------
        # INTERNAL USE ONLY
        #--------------------------------------------------------------------------------------------------------------

        iu = self._conf['internal parameters']

        self.interp_type = 'nearest'

        # self.bound_cell_clip_ratio = iu.get('boundary cell size cutoff', 0.5)

        self.substances_affect_charge = iu['substances affect Vmem']  # Do Network substances function bioelectrically?

         # default free diffusion constants (cytoplasmic)
        self.Do_Na = float(iu['Do_Na'])      # free diffusion constant sodium [m2/s]
        self.Do_K = float(iu['Do_K'])      # free diffusion constant potassium [m2/s]
        self.Do_Cl = float(iu['Do_Cl'])     # free diffusion constant chloride [m2/s]
        self.Do_Ca = float(iu['Do_Ca'])     # free diffusion constant calcium [m2/s]
        self.Do_M = float(iu['Do_M'])     # free diffusion constant mystery anchor ion [m2/s]
        self.Do_P = float(iu['Do_P'])      # free diffusion constant protein [m2/s]

        # fixed levels of ATP, ADP and Pi, required for use in pump equations to get propper kinetics
        self.cATP = 1.5
        self.cADP = 0.1
        self.cPi = 0.1

        # pump parameters
        self.alpha_NaK = float(iu['alpha_NaK']) # maximum rate constant sodium-potassium ATPase per unit surface area

        # FIXME add these as options to the config
        self.KmNK_Na = 12.0   # NaKATPase enzyme ext Na half-max sat value (alpha1 = 12, alpha2 = 20, alpha3 = 60)
        self.KmNK_K = 0.2     # NaKATPase enzyme ext K half-max sat value (alpha1 = 0.2, alpha2 = 0.20, alpha3 = 0.09)
        self.KmNK_ATP = 0.5   # NaKATPase enzyme ATP half-max sat value

        self.alpha_Ca = float(iu['alpha_Ca']) # pump rate for calcium ATPase in membrane [1/mol*s] 2.0e-15

        # FIXME add these as options to the config:
        self.KmCa_Ca = 1.0e-3   # CaATPase enzyme Ca half-max sat value (1.7 - 2.8 for vascular, 0.25 for platlets)
        self.KmCa_ATP = 0.5    # CaATPase enzyme ATP half-max sat value

        # partial pressure dissolved CO2
        self.CO2 = 40.0   # [mmHg]
        self.cCO2 = 1.2

        # charge states of ions
        self.z_Na = 1
        self.z_K = 1
        self.z_Cl = -1
        self.z_Ca = 2
        self.z_H = 1
        self.z_P = -1
        self.z_M = -1

        # molar mass of ions (kg/mol):
        self.M_Na = 23e-3
        self.M_K = 39e-3
        self.M_Cl = 35e-3
        self.M_Ca = 40e-3
        self.M_H = 1e-3
        self.M_P = 500e-3
        self.M_M = 60e-3

        # fundamental constants
        self.F = 96485 # Faraday constant [J/V*mol]
        self.R = 8.314  # Gas constant [J/K*mol]
        self.eo = 8.854e-12 # permittivity of free space [F/m]
        self.kb = 1.3806e-23  # Boltzmann constant [m2 kg/ s2 K1]
        self.q = 1.602e-19    # electron charge [C]
        self.mu = 1.275e-6   # magnetic permeability [H/m or N/A2]
        self.NAv = 6.022e23     # Avagadro's Number
        self.er = 80.0          # relative dielectric constant of electrolyte
        self.eedl = float(iu.get('dielectric constant', 40.0))         # relative dielectric constant of screening layer

        self.deltaGATP = -37000    # free energy released in ATP hydrolysis under standard phys conditions [J/mol]

        self.ac = 1.0e-6  # cell-cell separation for drawing
        self.scale_cell = 0.99          # the amount to scale cell membranes in from ecm edges (only affects drawing)
        self.cm = float(iu['membrane capacitance'])           # patch capacitance of cell membrane 0.022 [F/m2]

        self.tm = 7.5e-9           # thickness of cell membrane [m]
        self.cell_sides = 4      # minimum number of membrane domains per cell (must be >2)
        self.scale_alpha = 1.4   # the amount to scale (1/d_cell) when calculating the concave hull (boundary search)
        self.merge_cut_off = (1/50)  # the fraction of nominal cell perimeter at which nearby ecm points are merged

        self.d_cell = self.cell_radius * 2  # diameter of single cell
        self.nx = int(self.wsx / self.d_cell)  # number of lattice sites in world x index
        self.ny = int(self.wsy / self.d_cell)  # number of lattice sites in world y index
        self.wsx = self.wsx + 5 * self.cell_lattice_disorder * self.d_cell  # readjust the world size for noise
        self.wsy = self.wsy + 5 * self.cell_lattice_disorder * self.d_cell

        self.gjl = 2*self.tm + self.cell_space     # gap junction length
        self.gj_radius = 1.0e-9              # effective radius of gap junctions connecting cells [m] (range 0 to 5.0 e-9 m)

        self.um = 1e6    # multiplication factor to convert m to um

        self.isamples = 40.0  # sampling of vector data for currents

        self.rho = 1050 # mass density of system [kg/m3]

        self.fast_update_ecm = iu.get('fast update ecm', False)  # quick or slow update to cell<--> ecm grid exchange?

        self.sharpness = float(iu.get('sharpness env', 0.999))

        self.smooth_cells = 1/float(iu.get('sharpness cell', 0.5))

        self.true_cell_size = float(iu.get('true cell size', 1.0e-5))

        #FIXME: Can this initialization be safely moved earlier -- say, directly
        #*AFTER* tissue profile initialization required by this initialization?

        # Initialize the ion profile specified by this configuration.
        self._load_ion_profile()

        # Return this configuration for convenience.
        return self

    # ..................{ LOADERS                            }..................
    #FIXME: Ideally, this method should be private. Unfortunately, external
    #callers (notably, tests) currently need to call this method to update
    #absolute pathnames after modifying relative pathnames. The solution, of
    #course, is to have Python implicitly call this method whenever *ANY*
    #relative pathname internally referenced by this modified. Doing so
    #correctly will require augmenting the expr_alias() data descriptor to
    #support yet-another-optional-keyword-parameter enabling callers to pass a
    #callable to be implicitly called whenever that descriptor is set -- say,
    #"callback_set". Given that, we would then refactor the above YAML aliases
    #to resemble something like:
    #
    #    sim_pickle_basename = yaml_alias(
    #        "['sim file saving']['file']", str, callback_set=self.load_path)
    #
    #Oh... wait. Obviously, we have no access to "self.load_path" from class
    #scope. Err; perhaps refactor that to require a method accepting the current
    #"Parameters" object be passed, which would then permit convenient calling
    #via lambdas ala:
    #
    #    sim_pickle_basename = yaml_alias(
    #        "['sim file saving']['file']", str,
    #        callback_set: lambda p: p.load_path())
    #
    #Right. That's definitely it. Given that, we could then define and reuse a
    #trivial private function in this submodule resembling:
    #
    #    @type_check
    #    def _reload_paths(p: Parameters) -> none:
    #        p.reload_paths()
    #
    #Works for us. *shrug*
    def reload_paths(self) -> None:
        '''
        Redefine all absolute pathnames depending upon relative pathnames
        specified by this configuration.

        To avoid desynchronization between the two, this method *must* be called
        on every change to such a relative pathname.
        '''

        # Absolute dirnames to which phase results are pickled.
        self.init_pickle_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.init_pickle_dirname_relative)
        self.sim_pickle_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.sim_pickle_dirname_relative)

        # Absolute dirnames to which phase results are exported.
        self.init_export_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.init_export_dirname_relative)
        self.sim_export_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.sim_export_dirname_relative)

        # Absolute pathnames to which GRN results are pickled.
        self.grn_pickle_dirname = pathnames.join(
            self.conf_dirname, self.grn_pickle_dirname_relative)
        self.grn_pickle_filename = pathnames.join(
            self.grn_pickle_dirname, self.grn_pickle_basename)

        # Absolute pathnames from which prior GRN results are unpickled.
        self.grn_unpickle_filename = None
        if self.grn_unpickle_filename_relative is not None:
            # logs.log_debug('GRN load from: %s', self.grn_unpickle_filename_relative)
            self.grn_unpickle_filename = pathnames.join(
                self.conf_dirname, self.grn_unpickle_filename_relative)

        # Create all non-existing directories.
        dirs.make_unless_dir(
            self.init_pickle_dirname, self.sim_pickle_dirname,
            self.init_export_dirname, self.sim_export_dirname,
            self.grn_pickle_dirname,
        )


    def _load_ion_profile(self) -> None:
        '''
        Initialize the ion profile specified by this configuration.
        '''

        # Default tissue profile providing base membrane diffusion constants.
        base = self.tissue_default

        # simplest ion profile giving realistic results with minimal ions (Na+ & K+ focus):
        if self.ion_profile is IonProfileType.BASIC:
            self.cNa_env = 145.0
            self.cK_env = 5.0
            self.cP_env = 10.0

            zs = [self.z_Na, self.z_K, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 12.0
            self.cK_cell = 139.0
            self.cP_cell = 135.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cP_cell]

            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.ions_dict = {'Na':1, 'K':1, 'Cl':0, 'Ca':0, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'P': self.cP_cell, 'M': self.cM_cell}
            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'P': self.cP_env, 'M': self.conc_env_m}
            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'P': base.Dm_P, 'M': base.Dm_M}
            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'P': self.z_P, 'M': self.z_M}
            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'P': self.Do_P, 'M': self.Do_M}
            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'P': self.M_P, 'M': self.M_M}
            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'P':'proteins', 'M':'anion'}

        elif self.ion_profile is IonProfileType.BASIC_CA:
            self.cNa_env = 145.0
            self.cK_env = 5.0
            self.cCa_env = 2.0
            self.cP_env = 10.0

            zs = [self.z_Na, self.z_K, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 12.0
            self.cK_cell = 139.0
            self.cCa_cell = 1.0e-4
            self.cP_cell = 135.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cCa_cell, self.cP_cell]

            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.cCa_er = 0.5
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':0, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'P': self.cP_cell, 'M': self.cM_cell}
            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'P': self.cP_env, 'M': self.conc_env_m}
            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'P': base.Dm_P, 'M': base.Dm_M}
            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'P': self.z_P, 'M': self.z_M}
            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'P': self.Do_P, 'M': self.Do_M}
            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'P': self.M_P, 'M': self.M_M}
            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'P':'proteins', 'M':'anion'}

        # default environmental and cytoplasmic initial values mammalian cells
        elif self.ion_profile is IonProfileType.MAMMAL:
            # initialize proton concentrations to "None" placeholders
            self.cNa_env = 145.0
            self.cK_env = 5.0
            self.cCl_env = 115.0
            self.cCa_env = 2.0
            self.cP_env = 10.0

            zs = [self.z_Na, self.z_K, self.z_Cl, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCl_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 12.0
            self.cK_cell = 139.0
            self.cCl_cell = 4.0
            self.cCa_cell = 5.0e-5
            self.cP_cell = 135.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cCl_cell, self.cCa_cell, self.cP_cell]
            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.cCa_er = 0.5
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':1, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'Cl': self.cCl_cell,
                              'P': self.cP_cell, 'M': self.cM_cell}

            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'Cl': self.cCl_env,
                             'P': self.cP_env, 'M': self.conc_env_m}

            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'Cl': base.Dm_Cl,
                              'P': base.Dm_P, 'M': base.Dm_M}

            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'Cl': self.z_Cl,
                               'P': self.z_P, 'M': self.z_M}

            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'Cl': self.Do_Cl,
                              'P': self.Do_P, 'M': self.Do_M}

            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'Cl': self.M_Cl,
                               'P': self.M_P, 'M': self.M_M}

            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'Cl':'chloride',
                                  'P':'proteins', 'M':'anion'}

        elif self.ion_profile is IonProfileType.AMPHIBIAN:
            # initialize proton concentrations to "None" placeholders
            self.cNa_env = 14.50
            self.cK_env = 0.5
            self.cCl_env = 10.50
            self.cCa_env = 0.2
            self.cP_env = 0.0

            zs = [self.z_Na, self.z_K, self.z_Cl, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCl_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 8.0
            self.cK_cell = 125.0
            self.cCl_cell = 20.0
            self.cCa_cell = 1.0e-3
            self.cP_cell = 100.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cCl_cell, self.cCa_cell, self.cP_cell]
            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.cCa_er = 0.5
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':1, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'Cl': self.cCl_cell,
                              'P': self.cP_cell, 'M': self.cM_cell}
            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'Cl': self.cCl_env,
                             'P': self.cP_env, 'M': self.conc_env_m}
            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'Cl': base.Dm_Cl,
                              'P': base.Dm_P, 'M': base.Dm_M}
            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'Cl': self.z_Cl,
                               'P': self.z_P, 'M': self.z_M}
            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'Cl': self.Do_Cl,
                              'P': self.Do_P, 'M': self.Do_M}

            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'Cl': self.M_Cl,
                               'P': self.M_P, 'M': self.M_M}

            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'Cl':'chloride',
                                  'P':'proteins', 'M':'anion'}

        # user-specified environmental and cytoplasm values (customized)
        elif self.ion_profile is IonProfileType.CUSTOM:
            # initialize proton concentrations to "None" placeholders
            cip = self._conf['general options']['customized ion profile']

            self.cNa_env = float(cip['extracellular Na+ concentration'])
            self.cK_env = float(cip['extracellular K+ concentration'])
            self.cCl_env = float(cip['extracellular Cl- concentration'])
            self.cCa_env = float(cip['extracellular Ca2+ concentration'])
            self.cP_env = float(cip['extracellular protein- concentration'])

            zs = [self.z_Na, self.z_K, self.z_Cl, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCl_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env, zs)

            self.cNa_cell = float(cip['cytosolic Na+ concentration'])
            self.cK_cell = float(cip['cytosolic K+ concentration'])
            self.cCl_cell = float(cip['cytosolic Cl- concentration'])
            self.cCa_cell = float(cip['cytosolic Ca2+ concentration'])
            # self.cM_cell = float(cip['cytosolic HCO3- concentration'])
            self.cP_cell = float(cip['cytosolic protein- concentration'])

            conc_cell = [self.cNa_cell, self.cK_cell, self.cCl_cell, self.cCa_cell, self.cP_cell]
            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell, zs)

            assert self.z_M_cell == -1

            self.cCa_er = float(cip['endoplasmic reticulum Ca2+'])
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':1, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'Cl': self.cCl_cell, 'P': self.cP_cell, 'M': self.cM_cell}

            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'Cl': self.cCl_env, 'P': self.cP_env, 'M': self.conc_env_m}

            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'Cl': base.Dm_Cl, 'P': base.Dm_P, 'M': base.Dm_M}

            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'Cl': self.z_Cl, 'P': self.z_P, 'M': self.z_M}

            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'Cl': self.Do_Cl, 'P': self.Do_P, 'M': self.Do_M}

            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'Cl': self.M_Cl, 'P': self.M_P, 'M': self.M_M}

            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'Cl':'chloride',
                                  'P':'proteins', 'M':'anion'}

        # Else, this ion profile type is unrecognized. Raise an exception.
        else:
            raise BetseSimConfException(
                'Ion profile type "{}" unrecognized.'.format(self.ion_profile))

    # ..................{ UNLOADERS                          }..................
    def unload(self) -> None:

        # Unload our superclass.
        super().unload()

        # Unload all previously loaded network subconfigurations.
        self.grn.unload()

        # Unload all previously loaded tissue subconfigurations.
        self.cut_profiles.unload()
        self.tissue_default.unload()
        self.tissue_profiles.unload()

        # Unload all previously loaded export subconfigurations.
        self.anim.unload()
        self.csv.unload()
        self.plot.unload()

    # ..................{ EXCEPTIONS                         }..................
    def die_unless_ecm(self) -> None:
        '''
        Raise an exception unless this configuration has enabled simulation of
        the extracellular matrix and hence environmental grid spaces.
        '''

        if not self.is_ecm:
            raise BetseSimConfException(
                'Extracellular spaces disabled by '
                'this simulation configuration.')

    # ..................{ SETTERS                            }..................
    #FIXME: Actually, this method makes little sense here; clearly, this method
    #sets phase-specific attributes. All phase-specific attributes should reside
    #directly in the "SimPhase" API rather than in this class, which is
    #supposed to be phase-agnostic and hence reusable between phases.

    @type_check
    def set_time_profile(self, phase_kind: SimPhaseKind) -> None:
        '''
        Set temporal attributes specific to the passed simulation phase type.

        These attributes include:

        * ``dt``, the duration in seconds of each time step for this phase.
        * ``total_time``, the duration in seconds of this phase.

        Parameters
        ----------
        phase_kind : SimPhaseKind
            Current simulation phase type.
        '''

        # If this simulation phase is an initialization...
        if phase_kind is SimPhaseKind.INIT:
            self.dt = self.init_time_step
            self.init_tsteps = self.init_time_total / self.dt
            self.resample = self.init_time_sampling
            self.total_time = self.init_time_total
        # Else if this simulation phase is a simulation.
        elif phase_kind is SimPhaseKind.SIM:
            self.dt = self.sim_time_step
            self.sim_tsteps = self.sim_time_total / self.dt
            self.resample = self.sim_time_sampling
            self.total_time = self.sim_time_total
        else:
            raise BetseSimPhaseException(
                'Simulation phase "{}" unsupported.'.format(phase_kind.name))

        # Number of time steps (including sampled and unsampled) between each
        # unsampled time step, including that unsampled time step itself.
        self.t_resample = self.resample / self.dt

        # Duration in seconds of the current simulation phase accelerated by
        # the current gap junction acceleration factor.
        self.total_time_accelerated = self.total_time

    # ..................{ SUPERCLASS                         }..................
    #FIXME: Actually implement this properly. Ideally, this method should return
    #the set of all subdirectories internally referenced by the current
    #top-level YAML file. Instead, it simply returns all subdirectories
    #internally referenced by the *DEFAULT* top-level YAML file. Why? Because
    #the latter was much easier than the former, despite being wrong. Laziness
    #prevails!
    def _iter_conf_subdirnames(self) -> IterableTypes:

        # If no YAML file has been loaded yet, raise an exception.
        self.die_unless_loaded()

        # Default to the basenames of all direct subdirectories of the parent
        # directory containing the default simulation configuration file, which
        # are guaranteed to be required by this file.
        return dirs.iter_subdir_basenames(pathtree.get_data_yaml_dirname())

# ....................{ HELPERS                            }....................
def _balance_charge(concentrations: SequenceTypes, zs: SequenceTypes) -> tuple:
    '''
    Sum the concentrations of profile ions with their charge state to
    determine how much net positive charge exists, returning the concentration
    of the charge compensation anion M- needed to have zero net charge.

    Parameters
    -------------
    concentrations : SequenceType
        Array defining the concentrations of all ions in a space.
    zs : SequenceType
        Array (in complementary order to `concentrations`) of ion valence
        state.

    Returns
    ---------
    (float, float)
        2-tuple `(bal_conc, valence)`, where:
        * `bal_conc` is the concentration of anion M- to create zero net
          charge.
        * `valence` is the charge of the `bal_conc`. Ideally, this should
          _always_ be -1.
    '''

    q = 0

    for conc,z in zip(concentrations, zs):
        q = q+ conc*z

        to_zero = -q
        bal_conc = abs(to_zero)
        valance = np.sign(to_zero)

        assert bal_conc >= 0

    return bal_conc, valance



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level module facilities.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory dependencies,
# the top-level of this module may import *ONLY* from packages guaranteed to
# exist at installation time -- which typically means *ONLY* BETSE packages and
# stock Python packages.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import importlib, sys
from betse.exceptions import BetseModuleException
from betse.util.io.log import logs
from betse.util.type import types
from betse.util.type.mapping.mapcls import DefaultDict
from betse.util.type.types import (
    type_check,
    ModuleType,
    ModuleOrStrTypes,
    SetType,
    StrOrNoneTypes,
)
from collections import defaultdict
from importlib import util as importlib_util
from importlib.machinery import ExtensionFileLoader, EXTENSION_SUFFIXES

# ....................{ GLOBALS ~ dict                     }....................
DISTUTILS_PROJECT_NAME_TO_MODULE_NAME = DefaultDict(
    missing_key_value=lambda self, missing_key: missing_key,
    initial_mapping={
        'Numpy': 'numpy',
        'Pillow': 'PIL',
        'PyYAML': 'yaml',
        'SciPy': 'scipy',
    }
)
'''
Dictionary mapping a relevant :mod:`distutils`-specific project name (e.g.,
``PyYAML``) to the fully-qualified name of the corresponding module or package
providing that project (e.g., ``yaml``).

All modules and packages explicitly unmapped by this dictionary default to the
identity mapping -- that is, mapping the fully-qualified names of those modules
and packages to themselves (e.g., from ``pytest`` to ``pytest``).
'''


MODULE_NAME_TO_VERSION_ATTR_NAME = defaultdict(
    # Default attribute name to be returned for all unmapped modules.
    lambda: '__version__',

    # Module-specific attribute names.
    PIL='PILLOW_VERSION',
)
'''
Dictionary mapping the fully-qualified name of a module or package (e.g.,
:mod:``PIL``) to the name of the attribute (e.g., ``PILLOW_VERSION``) declared
by that module or package, providing that module or package's version specifier.

All modules and packages explicitly unmapped by this dictionary default to the
canonical ``__version__`` attribute name.
'''

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_unless_module(
    module_name: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a message
    synthesized from the passed module name) if the module with the passed name
    is *not* importable by the active Python interpreter.

    Raises
    ----------
    ImportError
        If this module is unimportable. To permit callers to transparently
        handle importation errors in the standard way, this standard exception
        rather than an application-specific exception (e.g.,
        :class:`betse.exceptions.BetseModuleException`) is raised.

    See Also
    ----------
    :func:`is_module`
        Further details.
    '''

    # If this module is unimportable, raise an exception.
    if not is_module(module_name):
        # If no exception message was passed, synthesize one from this name.
        if not exception_message:
            exception_message = 'Module "{}" not found.'.format(module_name)

        # Raise this exception. (See the docstring for further details.)
        raise ImportError(exception_message)

# ....................{ TESTERS                            }....................
@type_check
def is_module(module_name: str) -> bool:
    '''
    ``True`` only if the module with the passed fully-qualified name is
    importable under the active Python interpreter.

    If this module is a **submodule** (i.e., contains a ``.`` character), all
    parent modules of this module will be imported as a side effect of this
    function call. Likewise, if this module is *not* importable via standard
    mechanisms (e.g., the OS X-specific :mod:`PyObjCTools` package), this module
    itself may also be imported as a side effect.
    '''

    # Depending on context, this function behaves in one of three distinct ways:
    #
    # * If this module's name is a key in the canonical dictionary "sys.modules"
    #   and has thus already been imported at least once under the active Python
    #   process, then...
    #   * If the "sys.modules[module_name].__spec__" attribute is set to a non-
    #     None value, that value is returned.
    #   * Else, the "ValueError" exception is raised.
    # * Else if this module is loadable by iteratively querying all module
    #   loaders in "sys.meta_path" (the canonical list of such loaders), a new
    #   spec is created describing this module and returned.
    # * Else this module is unloadable. In this case:
    #   * If this module is a submodule (i.e., this module's name contains a
    #     "." delimiter) and any parent module of this submodule is unloadable,
    #     the "ImportError" exception is raised.
    #   * Else, None is returned.
    #
    # Since this function only returns a single boolean, these return values and
    # exceptions are converted to simple boolean values.
    try:
        return importlib_util.find_spec(module_name) is not None
    # If this module is a submodule (i.e., this module's name contains a "."
    # delimiter) and any parent module of this submodule is unloadable, this
    # submodule itself is unloadable.
    except ImportError:
        return False
    # If this module appears to have been imported at least once under the
    # active Python process but has no "__spec__" attribute, inspect deeper.
    # This exception does *NOT* necessarily imply this module to not exist. This
    # module may exist even if this exception is thrown -- namely for modules
    # dynamically defined at runtime rather than loaded from external files.
    #
    # Unfortunately, this exception does imply that conventional alternatives to
    # the prior function call (e.g., testing tuples generated by
    # pkgutil.iter_modules()) will also fail to find this module. As a fallback,
    # attempt to manually import this module. Since doing so implicitly imports
    # the "__init__.py" files of all parent packages of this module and hence
    # may have unhelpful side effects, we do so only if the prior call failed.
    except ValueError:
        try:
            importlib.import_module(module_name)
            return True
        except ImportError:
            return False


@type_check
def is_imported(*module_names: str) -> bool:
    '''
    ``True`` only if all modules with the passed fully-qualified names have
    already been imported in the active Python process.
    '''

    # all(). It is awesome.
    return all(module_name in sys.modules for module_name in module_names)

# ....................{ TESTERS ~ type                     }....................
#FIXME: Add unit tests, as this is a fairly fragile tester.
@type_check
def is_c_extension(module: ModuleOrStrTypes) -> bool:
    '''
    ``True`` only if the passed module is a C extension implemented as a
    dynamically linked shared library specific to the current platform.

    Parameters
    ----------
    module : ModuleOrStrTypes
        Either:
        * The fully-qualified name of this module, in which case this function
          dynamically imports this module.
        * A previously imported module object.

    Returns
    ----------
    bool
        ``True`` only if this module is a C extension.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Resolve this module's object.
    module = _resolve_module(module)

    # If this module was loaded by a PEP 302-compliant C extension loader, this
    # module *MUST* be a C extension.
    if isinstance(getattr(module, '__loader__', None), ExtensionFileLoader):
        return True

    # Else, fallback to filetype matching heuristics.
    #
    # Absolute path of the file defining this module.
    module_filename = get_filename(module)

    # "."-prefixed filetype of this path if any or "None" otherwise.
    module_filetype = pathnames.get_filetype_dotted_or_none(module_filename)
    # print('module_filetype: {}'.format(module_filetype))

    #FIXME: Mildly inefficient, as "EXTENSION_SUFFIXES" is a list rather than a
    #set. Since this list is small *AND* since this function is called
    #infrequently, this is currently ignorable. The trivial fix is to define a
    #new private "_EXTENSION_SUFFIXES = set(EXTENSION_SUFFIXES)" global above
    #and leverage that here instead.

    # This module is only a C extension if this path's filetype is that of a
    # C extension specific to the current platform.
    return module_filetype in EXTENSION_SUFFIXES

# ....................{ GETTERS ~ path                     }....................
@type_check
def get_dirname(module: ModuleOrStrTypes) -> str:
    '''
    Absolute path of the directory containing the passed module.

    If this module is a non-namespace package, this is the directory containing
    this package's top-level ``__init__`` submodule.

    Parameters
    ----------
    module : ModuleOrStrTypes
        Either:
        * The fully-qualified name of this module, in which case this function
          dynamically imports this module.
        * A previously imported module object.

    Returns
    ----------
    str
        Absolute path of this directory.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Return this dirname.
    return pathnames.get_dirname(get_filename(module))


#FIXME: The current approach is trivial and therefore terrible, breaking down
#under commonplace real-world conditions (e.g., modules embedded within egg-like
#archives). Consider generalizing this approach via the new setuptools-based
#"betse.lib.setuptool.resources" submodule.
@type_check
def get_filename(module: ModuleOrStrTypes) -> str:
    '''
    Absolute path of the file providing the passed module or package.

    If the passed object signifies:

    * A package (e.g., directory), this is the absolute path of the file
      providing this package's ``__init__`` submodule.
    * A non-package (e.g., module, C extension), this is the absolute path of
      the file providing this non-package as is.

    Caveats
    ----------
    Since effectively *all* modules and packages (except in-memory builtin
    modules) are associated with a corresponding file, there intentionally
    exists no corresponding ``get_filename_or_none()`` function.

    Since *only* packages define the ``__path__`` attribute, there likewise
    exists no corresponding ``get_pathname()`` function. Only the ``__file__``
    attribute retrieved by this function is generally applicable to *all*
    modules and packages regardless of type.

    Parameters
    ----------
    module : str or ModuleType
        Either:
        * The fully-qualified name of this module, in which case this function
          dynamically imports this module.
        * A previously imported module object.

    Returns
    ----------
    str
        Absolute path of this file.

    Raises
    ----------
    BetseModuleException
        If this module has no such attribute (e.g., is a builtin module).
    '''

    # Resolve this module's object.
    module = _resolve_module(module)

    # If this module does *NOT* provide the special "__file__" attribute, raise
    # an exception. (All modules *EXCEPT* builtin modules should provide this.)
    if not hasattr(module, '__file__'):
        raise BetseModuleException(
            'Module "{0}.__file__" attribute not found '
            '(e.g., as "{0}" is a builtin module).'.format(module.__name__))

    # Else, return this attribute's value.
    return module.__file__

# ....................{ GETTERS ~ global                   }....................
@type_check
def get_global_names(module: ModuleOrStrTypes) -> SetType:
    '''
    Set of the names of all global variables defined by the passed module.

    This function returns the set of the names of all attributes defined by this
    module, excluding:

    * Special attributes reserved for use by Python (e.g., ``__file__``).
    * Callable attributes (e.g., functions, lambdas).

    Parameters
    ----------
    module : ModuleOrStrTypes
        Either:
        * The fully-qualified name of this module, in which case this function
          dynamically imports this module.
        * A previously imported module object.

    Returns
    ----------
    set
        Set of the names of all global variables defined by this module.
    '''

    # Resolve this module's object.
    module = _resolve_module(module)

    # Return this set via a set comprehension.
    return {
        module_attr_name
        for module_attr_name in dir(module)
        if (
            (not module_attr_name.startswith('__')) and
            callable(getattr(module, module_attr_name))
        )
    }

# ....................{ GETTERS ~ version                  }....................
@type_check
def get_version(module: ModuleOrStrTypes) -> str:
    '''
    Version specifier of the passed module.

    If this module provides no version specifier, an exception is raised.

    See Also
    ----------
    :func:`get_version_or_none`
        Further details on the passed parameter.
    '''

    # Module version if any or "None" otherwise.
    module_version = get_version_or_none(module)

    # If this version does *NOT* exist, raise an exception.
    if module_version is None:
        raise BetseModuleException(
            'Module "{}" version not found.'.format(module))

    # Return this version.
    return module_version


@type_check
def get_version_or_none(module: ModuleOrStrTypes) -> StrOrNoneTypes:
    '''
    Version specifier of the passed module if that module provides a version
    specifier *or* ``None`` otherwise.

    Parameters
    ----------
    module : ModuleOrStrTypes
        Either:
        * The fully-qualified name of this module, in which case this function
          dynamically imports this module.
        * A previously imported module object.

    Returns
    ----------
    StrOrNoneTypes
        This module's version specifier if any *or* ``None`` otherwise.
    '''

    # Resolve this module's object.
    module = _resolve_module(module)

    # Name of the version specifier attribute defined by that module. For sane
    # modules, this is "__version__". Insane modules, however, exist.
    module_version_attr_name = MODULE_NAME_TO_VERSION_ATTR_NAME[module.__name__]

    # This attribute defined by this module if any or "None" otherwise.
    module_version = getattr(module, module_version_attr_name, None)

    # If this version is undefined, log a non-fatal warning.
    if module_version is None:
        logs.log_warning('Module "%s" version not found.', module.__name__)

    # Return this version.
    return module_version

# ....................{ IMPORTERS                          }....................
@type_check
def import_module(
    module_name: str, exception_message: StrOrNoneTypes = None) -> ModuleType:
    '''
    Dynamically import and return the module, package, or C extension with the
    passed fully-qualified name if importable *or* raise an exception with the
    passed message otherwise.
    '''

    # If this module is unimportable, raise an exception.
    die_unless_module(module_name, exception_message)

    # Else, import and return this module.
    return importlib.import_module(module_name)


@type_check
def unimport_module_if_imported(*module_names: str) -> None:
    '''
    Dynamically unimport each of the modules, packages, or C extensions with the
    passed fully-qualified names that have been previously imported under the
    active Python interpreter.

    Specifically, for each passed name:

    * If this name is that of a previously imported module, this module is
      removed from the canonical :attr:`sys.modules` cache of all previously
      imported modules. On the next attempt to import this module, Python will
      re-import and re-cache that module to this cache.
    * Else, this module is silently ignored.
    '''

    # For each of the passed module names...
    for module_name in module_names:
        # If this module has already been imported, unimport this module.
        if is_imported(module_name):
            logs.log_debug('Unimporting module "{}"...'.format(module_name))
            del sys.modules[module_name]
        # Else, this module has *NOT* yet been imported. Ignore this module.

# ....................{ PRIVATE ~ resolvers                }....................
@type_check
def _resolve_module(module : ModuleOrStrTypes) -> ModuleType:
    '''
    Dynamically import and return the module with the passed name if a string
    is passed *or* return the passed module as is otherwise.

    This utility function is intended *only* to simplify the implementation of
    public functions defined by this submodule and is hence private.

    Parameters
    ----------
    module : ModuleOrStrTypes
        Either:
        * The fully-qualified name of this module, in which case this function
          dynamically imports this module.
        * A previously imported module object.

    Returns
    ----------
    ModuleType
        Module object resolved from the passed parameter.
    '''

    # If a module name was passed, dynamically import and return this module.
    if types.is_str(module):
        return import_module(module)
    # Else, type checking guarantees this object to be a module. Return this.
    else:
        return module



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level application initialization common to both the CLI and GUI.
'''

#FIXME: Raise an exception if running with superuser privelages. To avoid
#inevitable security issues, Under *NO* circumstances should either BETSE or
#BETSEE ever be run with elevated permissions. To do so, see the following
#canonical answer:
#    https://stackoverflow.com/a/1026626/2809027

#FIXME: Print memory-related metadata when running "betse info" *AND* log
#non-fatal warnings when BETSE is run under a low-memory environment (e.g.,
#< 4GB available free memory). To do so, note the following canonical API:
#
#    psutil.Process(os.getpid()).get_memory_info()

#FIXME: Consider replacing bottleneck Numpy routines with routines imported from
#the following third-party Numpy-like frameworks:
#
#* "bottleneck", providing optimized routines accepting Numpy arrays --
#  implemented in Cython and hence faster than comparible Numpy routines.
#* "numexpr", a Theano-like framework accepting Numpy arrays -- performing
#  CPU-centric parallelization of expensive array operations. Whereas Theano
#  permits such operations to be conveniently expressed in pure-Python, however,
#  numexpr inconveniently requires such operations be expressed as... wait for
#  it, raw strings. So, that sucks. Nonetheless, worth a possible look.
#* "blaze", a purported second-gen Numpy replacement. We harbour sincere doubts,
#  but everything deserves its millisecond to shine in the light. Ah. We see.
#  Blaze is considerably more heavy-weight than Numpy, and largely serves a
#  completely different marketshare: supercomputing. That's well beyond our
#  means, at the moment. Numpy it is!
#
#In short, "bottleneck" is probably the only framework listed above of interest.

#FIXME: Consider optimizing frequently used matrix and vector computations with
#Theano, a general-purpose Python mathematical optimization framework. One
#particularly compelling use case for Theano is to portably distribute
#computational work across multiple GPUs. In general, Theano can be used to
#reduce arbitrarily complex symbolic expressions expressed in pure Python to
#dynamically compiled machine code on-the-fly. Fairly amazing, all around. For
#the high-level synopsis, see:
#
#    http://deeplearning.net/software/theano/introduction.html
#FIXME: Theano and Torch (a similar framewark) appear to now be subsumed by
#TensorFlow, a Google-backed framework originally implemented in support of
#machine learning workflows at Google (e.g., DeepMind), but sufficiently
#generalized as to support a wide variety of computational needs -- like ours.

#FIXME: The "~/.betse" directory grows fairly large fairly quickly. It'd be
#great to emit non-fatal warnings if its size exceeds some reasonable threshold
#(e.g., 1MB).

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To defer heavyweight and possibly circular imports, the top-level of
# this module may import *ONLY* from standard Python packages. All imports from
# application and third-party packages should be deferred to their point of use.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ GLOBALS                            }....................
_IS_IGNITED = False
'''
``True`` only if the :func:`init` function has already been called.

That function uses this private boolean to guard against repeated invocations of
the :func:`init` function from multiple modules in the same Python process
(e.g., :mod:`betse.science.__init__`, :mod:`betse.util.cli.cliabc`). While that
function does technically support repeated calls, each additional call after the
first inefficiently performs no meaningful work and is thus safely ignorable.
'''

# ....................{ IGNITERS                           }....................
def ignite() -> None:
    '''
    Initialize both the current application *and* all mandatory third-party
    dependencies of this application with sane defaults.

    This high-level convenience function intentionally provides *no* means of
    initializing either this application or these dependencies with alternative
    parameters. To do so, callers should call all lower-level initialization
    functions directly (e.g., :func:`init`, :func:`betse.lib.libs.init`).
    '''

    # Defer heavyweight and possibly circular imports.
    from betse.lib import libs

    # Initialize this application.
    init()

    # Initialize all dependencies *AFTER* initializing this application.
    libs.init()

# ....................{ INITIALIZERS                       }....................
def reinit() -> None:
    '''
    (Re-)initialize this application -- but *not* mandatory third-party
    dependencies of this application, which requires external resources (e.g.,
    command-line options, configuration files) to be parsed.

    Specifically:

    * If this application has *not* already been initialized under the active
      Python process, this application will be initilialized.
    * Else, this application has already been initialized under the active
      Python process. In this case, this application will be re-initilialized.

    See Also
    ----------
    :func:`betse.lib.libs.reinit`
        Function (re)-initializing all mandatory third-party dependencies.
    '''

    # Force the init() function to reinitialize this application.
    global _IS_IGNITED
    _IS_IGNITED = False

    # Reinitialize this application.
    init()


def init() -> None:
    '''
    Initialize the current application if this application has not already been
    initialized under the active Python process *or* noop otherwise.

    Specifically, this function (in order):

    #. Enables Python's standard handler for segmentation faults.
    #. Enables this application's default logging configuration.
    #. Validates (but does *not* initialize) all mandatory third-party
       dependencies of this application, which the :func:`betsee.lib.libs.init`
       function initializes independently.
    #. Validates core directories and files required at program startup,
       creating all such directories and files that do *not* already exist and
       are reasonably creatable.
    #. Validates the active Python interpreter (e.g., to support
       multithreading).

    To support caller-specific error handling, this function is intended to be
    called immediately *after* this application begins catching otherwise
    uncaught exceptions.
    '''

    # If this function has already been called, noop.
    global _IS_IGNITED
    if     _IS_IGNITED:
        return

    # Defer heavyweight and possibly circular imports.
    from betse.lib import libs
    from betse.util.io.error import errfault
    from betse.util.io.log import logconfig
    from betse.util.py import pys

    # Enable Python's standard handler for segmentation faults *BEFORE*
    # performing any further logic, any of which could conceivably trigger a
    # segmentation fault and hence process termination.
    errfault.handle_faults()

    # Enable the default logging configuration for the current Python process
    # *BEFORE* performing any validation, thus logging any exceptions raised by
    # this validation.
    logconfig.init()

    # Validate mandatory dependencies. Avoid initializing these dependencies
    # here (e.g., by calling libs.init()), which requires the logging
    # configuration to have been finalized (e.g., by parsing CLI options), which
    # has yet to occur this early in the application lifecycle.
    libs.die_unless_runtime_mandatory_all()

    # Validate the active Python interpreter *AFTER* mandatory dependencies.
    # While the former (mostly) comprises unenforced recommendations, the latter
    # comprises enforced requirements and should thus be validated first.
    pys.init()

    # Record this function as having been called *AFTER* successfully doing so.
    _IS_IGNITED = True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Decorators skipping tests.

These decorators conditionally mark their decorated tests as skipped depending
on whether the conditions signified by the passed parameters are satisfied
(e.g., the importability of the passed module name).
'''

# ....................{ IMPORTS                            }....................
import pytest
from betse.util.type.types import (
    type_check,
    CallableTypes,
    ClassType,
    SequenceOrNoneTypes,
    MappingOrNoneTypes,
)

# ....................{ IMPORTS ~ private                  }....................
# Sadly, the following imports require private modules and packages.
from _pytest.runner import Skipped

# ....................{ SKIP                               }....................
skip_if = pytest.mark.skipif
'''
Conditionally skip the decorated test with the passed human-readable
justification if the passed boolean is `False`.

Parameters
----------
boolean : bool
    Boolean to be tested.
reason : str
    Human-readable message justifying the skipping of this test.
'''


@type_check
def skip(reason: str):
    '''
    Unconditionally skip the decorated test with the passed human-readable
    justification.

    This decorator is intended to be called both directly as a function *and*
    indirectly as a decorator, which differs from both:

    * :func:`pytest.skip`, intended to be called only directly as a function.
      Attempting to call this function indirectly as a decorator produces
      extraneous ignorable messages on standard output resembling
      ``"SKIP [1] betse_test/unit/test_import.py:66: could not import 'xdist'"``,
      for unknown (and probably uninteresting) reasons.
    * :func:`pytest.mark.skip`, intended to be called only indirectly as a
      decorator. Attempting to call this decorator directly as a function
      reduces to a noop, for unknown (and probably uninteresting) reasons.

    Parameters
    ----------
    reason : str
        Human-readable message justifying the skipping of this test.
    '''

    return skip_if(True, reason=reason)

# ....................{ SKIP ~ command                     }....................
@type_check
def skip_unless_command(pathname: str):
    '''
    Skip the decorated test unless a command with the passed path exists.

    This is the case if this path is neither:

    * The basename of an executable file in the current ``${PATH}``.
    * The relative or absolute path of an executable file.

    This decorator is typically applied to tests requiring optional third-party
    external commands.

    Parameters
    ----------
    pathname : str
        Basename or absolute or relative path of the executable file to inspect.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator successfully reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.util.path.command import cmds
    from betse.util.type.decorator.decorators import decorator_identity

    # If this command exists, reduce this decoration to a noop.
    if cmds.is_command(pathname):
        return decorator_identity
    # Else, skip this test with a human-readable justification.
    else:
        return skip('Command "{}" not found.'.format(pathname))

# ....................{ SKIP ~ lib                         }....................
@type_check
def skip_unless_matplotlib_anim_writer(writer_name: str):
    '''
    Skip the decorated test unless the external command underlying the
    matplotlib animation writer with the passed name is in the current
    ``${PATH}`` (e.g., for the ``imagemagick`` writer, if the ``convert``
    command  is found).

    Parameters
    ----------
    writer_name : str
        Name of the matplotlib animation writer to be inspected.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator successfully reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.lib.matplotlib.writer import mplvideo
    from betse.util.type.decorator.decorators import decorator_identity

    # If this command exists, reduce this decoration to a noop.
    if mplvideo.is_writer(writer_name):
        return decorator_identity
    # Else, skip this test with a human-readable justification.
    else:
        return skip(
            'Matplotlib animation writer "{}" either not found or '
            'unrecognized by BETSE.'.format(writer_name))

# ....................{ SKIP ~ module : setuptools         }....................
#FIXME: Code duplication is bad. Let's stop replicating this structure.
@type_check
def skip_unless_lib_runtime_optional(*lib_names: str):
    '''
    Skip the decorated test if one or more of the optional runtime dependencies
    of this application with the passed :mod:`setuptools`-specific project names
    are **unsatisfiable** (i.e., unimportable *or* of unsatisfactory version).

    Parameters
    ----------
    lib_names : str
        Tuple of the names of all :mod:`setuptools`-specific projects
        identifying these dependencies (e.g., ``NetworkX``).

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseLibException
    from betse.lib import libs

    # Skip this test if one or more such dependences are unsatisfiable.
    return _skip_if_callable_raises_exception(
        exception_type=BetseLibException,
        func=libs.die_unless_runtime_optional,
        args=lib_names,
    )


@type_check
def skip_unless_requirement(*requirement_strs: str):
    '''
    Skip the decorated test if one or more of the dependencies identified by the
    passed :mod:`setuptools`-formatted requirement strings are **unsatisfiable**
    (i.e., unimportable *or* of unsatisfactory version).

    Parameters
    ----------
    requirement_strs : str
        Tuple of all :mod:`setuptools`-formatted requirement strings
        identifying these dependencies (e.g., `Numpy >= 1.8.0`).

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseLibException
    from betse.lib.setuptools import setuptool

    # Skip this test if one or more such dependences are unsatisfiable.
    return _skip_if_callable_raises_exception(
        exception_type=BetseLibException,
        func=setuptool.die_unless_requirement_str,
        args=requirement_strs,
    )

# ....................{ SKIP ~ module : importlib          }....................
@type_check
def skip_unless_module(module_name: str, minimum_version: str = None):
    '''
    Skip the decorated test if the module with the passed name is unimportable
    *or* importable but of a version less than the passed minimum version if
    non-``None``.

    Note that tests requiring optional third-party Python dependencies should
    call the higher-level :func:`skip_unless_lib_runtime_optional` decorator
    instead, which implicitly validates the versions of those dependencies
    rather than requiring those versions be explicitly passed.

    Caveats
    ----------
    The higher-level :func:`skip_unless_lib_runtime_optional` and
    :func:`skip_unless_requirement` decorators (which are implemented in terms
    of robust :mod:`setuptools` machinery) should typically be called in lieu
    of this lower-level decorator (which are implemented in terms of fragile
    :mod:`importlib` machinery).

    See Also
    ----------
    :func:`skip_unless_lib_runtime_optional`
        Higher-level decorator skipping the decorated test if the passed
        optional runtime dependency is .

        Higher-level decorator skipping the decorated test if one or more passed
        requirement strings are unsatisfiable.

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be tested for.
    minversion : optional[str]
        Optional minimum version of this module as a dot-delimited string (e.g.,
        ``0.4.0``) to be tested for if any *or* ``None`` otherwise, in which
        case any version is acceptable. Defaults to ``None``.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator reducing to a noop otherwise.
    '''

    return _skip_if_callable_raises_exception(
        exception_type=Skipped,
        func=pytest.importorskip,
        args=(module_name, minimum_version),
    )

# ....................{ SKIP ~ plugin                      }....................
def skip_unless_plugin_xdist():
    '''
    Skip the decorated test if the ``pytest-xdist`` plugin is *not* installed.

    This decorator is typically applied to tests requiring **process isolation**
    (i.e., isolating tests to dedicated subprocesses of the current test
    session).  While this plugin provides such isolation out-of-the-box, vanilla
    :mod:`pytest` does not. Hence, these tests *must* be skipped in the absence
    of this plugin.

    Examples of tests requiring process isolation include:

    * Unit tests testing importability. Since Python caches imports performed by
      the active Python interpreter (e.g., via :attr:`sys.modules`) *and* since
      the order in which :mod:`pytest` runs tests should be assumed to be
      non-deterministic, importability *cannot* be reliably tested within a
      single Python process.
    '''

    return skip_unless_module('xdist')

# ....................{ PRIVATE ~ skip                     }....................
@type_check
def _skip_if_callable_raises_exception(
    # Mandatory parameters.
    exception_type: ClassType,
    func: CallableTypes,

    # Optional parameters.
    args: SequenceOrNoneTypes = None,
    kwargs: MappingOrNoneTypes = None,
):
    '''
    Skip the decorated test if calling the passed function with the passed
    positional and keyword arguments raises an exception of the passed type.

    If calling this function raises:

    * Any other type of exception, this test is marked as a failure.
    * No exception, this test continues as expected.

    Parameters
    ----------
    exception_type : ClassType
        Type of exception expected to be raised by this callable.
    func : CallableTypes
        Callable to pass these arguments.
    args : SequenceOrNoneTypes
        Sequence of all positional arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.
    kwargs : MappingOrNoneTypes
        Dictionary of all keyword arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.

    Returns
    ----------
    pytest.skipif
        Decorator skipping this test if this callable raises this exception *or*
        the identity decorator reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.util.type.decorator.decorators import decorator_identity

    # Default all unpassed arguments to sane values.
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}

    # Attempt to call this callable with these arguments.
    try:
        func(*args, **kwargs)
    # If this callable raises an expected exception, skip this test.
    except exception_type as exception:
        return skip(str(exception))
    # If this callable raises an unexpected exception, fail this test. To
    # preserve this exception's stack trace, reraise this exception as is.
    except Exception as exception:
        raise
        # # Print this exception's stacktrace to stderr.
        # stderrs.output_exception(heading=(
        #     'skip_unless_lib_runtime_optional{} '
        #     'raised unexpected exception:\n'.format(lib_names)))
        #
        # # Skip this test with this exception's message.
        # return skip(str(exception))
    # Else, this callable raised no exception. Reduce this decoration to a noop.
    else:
        return decorator_identity



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

"""
Controls a gene regulatory network.

Creates and electrodiffuses a suite of customizable general gene products in the
BETSE ecosystem, where the gene products are assumed to activate and/or inhibit
the expression of other genes (and therefore the production of other gene
products) in the gene regulatory network (GRN).
"""

import numpy as np
from betse.science import filehandling as fh
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.science.chemistry.networks import MasterOfNetworks
from betse.science.chemistry.netplot import set_net_opts
# from betse.science.math.systems import SimMaster


class MasterOfMolecules(object):

    def __init__(self, p):

        #FIXME: Extract the "GeneralNetwork.betse" basename into a new
        #configuration option, perhaps in the "init file saving" section.
        #FIXME: Replace "GeneralNetwork.betse" with "GeneralNetwork.betse.gz" to
        #compress this pickled file.

        # Define data paths for saving an initialization and simulation run:
        self.savedMoM = pathnames.join(p.init_pickle_dirname, 'GeneralNetwork.betse')

    def read_mol_config(self, sim, cells, p):

        # read the config file into a dictionary:
        config_dic = p.network_config

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        # initialize the substances of metabolism in a core field encapsulating
        # Master of Molecules:
        self.core = MasterOfNetworks(sim, cells, substances_config, p)

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # read in substance properties from the config file, and initialize basic properties:
        self.core.read_substances(sim, cells, substances_config, p)
        self.core.tissue_init(sim, cells, substances_config, p)


        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, sim, cells, p)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True

        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, sim, cells, p)
            self.channels = True

        else:
            self.channels = False

        # initialize modulators, if desired:

        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

        # read in network plotting options:
        self.core.net_plot_opts = config_dic.get('network plotting', None)

        # set plotting options for the network:
        set_net_opts(self.core, self.core.net_plot_opts, p)

        # plot up direction surfaces:  # FIXME move this to its own command in Simrunner
        # dyna_surf = SimMaster(config_dic, p)

        # after primary initialization, check and see if optimization required:
        opti = config_dic['optimization']['optimize network']
        self.core.opti_N = config_dic['optimization']['optimization steps']
        self.core.opti_method = config_dic['optimization']['optimization method']
        self.core.target_vmem = float(config_dic['optimization']['target Vmem'])
        self.core.opti_T = float(config_dic['optimization']['optimization T'])
        self.core.opti_step = float(config_dic['optimization']['optimization step'])

        if opti is True:
            logs.log_info("The General Network is being analyzed for optimal rates...")
            self.core.optimizer(sim, cells, p)
            self.reinitialize(sim, cells, p)

    #FIXME: Most of this method appears to have been copy-and-pasted
    #from the read_mol_config() method above. That's... not the best. Let's
    #extract the code shared in common between these two methods into a new
    #_init_mols() method internally called by these two methods.
    def reinitialize(self, sim, cells, p):

        # read the config file into a dictionary:
        config_dic = p.network_config

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        self.core.tissue_init(sim, cells, substances_config, p)

        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, sim, cells, p)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True

        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, sim, cells, p)
            self.channels = True

        else:
            self.channels = False

        # initialize modulators, if desired:
        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

    def run_core_sim(self, sim, cells, p):

        sim.vm = -50e-3*np.ones(sim.mdl)

        sim.time = []


        # initialize key fields of simulator required to interface (dummy init)
        sim.rho_pump = 1.0
        sim.rho_channel = 1.0

        # set molecules to not affect charge for sim-grn test-drives:
        p.substances_affect_charge = False

        # specify a time vector
        loop_time_step_max = p.init_tsteps
        # Maximum number of seconds simulated by the current run.
        loop_seconds_max = loop_time_step_max * p.dt
        # Time-steps vector appropriate for the current run.
        tt = np.linspace(0, loop_seconds_max, loop_time_step_max)

        tsamples = set()
        i = 0
        while i < len(tt) - p.t_resample:
            i = int(i + p.t_resample)
            tsamples.add(tt[i])

        self.core.clear_cache()
        self.time = []

        for t in tt:

            if self.transporters:
                self.core.run_loop_transporters(t, sim, cells, p)

            # if self.modulators:
            #     self.core.run_loop_modulators(sim, self.core, cells, p)

            self.core.run_loop(t, sim, cells, p)




            if t in tsamples:
                sim.time.append(t)

                logs.log_info('------------------' + str(np.round(t,3)) +' s --------------------')
                self.time.append(t)
                self.core.write_data(sim, cells, p)
                self.core.report(sim, p)

        logs.log_info('Saving simulation...')
        datadump = [self, cells, p]
        fh.saveSim(p.grn_pickle_filename, datadump)
        self.core.init_saving(cells, p, plot_type='init', nested_folder_name='Gene')
        self.core.export_eval_strings(p)
        self.core.export_equations(p)
        message = 'Gene regulatory network simulation saved to' + ' ' + p.grn_pickle_filename
        logs.log_info(message)

        logs.log_info('-------------------Simulation Complete!-----------------------')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulation phase** (e.g., cell cluster seeding, initialization, and
simulation) functionality.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseSimPhaseException
from betse.science.phase import phasecallbacks
from betse.science.phase.phasecallbacks import (
    SimCallbacksABCOrNoneTypes, SimCallbacksNoop)
from betse.science.phase.phaseenum import SimPhaseKind
from betse.util.type.types import type_check, NoneType

# ....................{ CLASSES                            }....................
class SimPhase(object):
    '''
    High-level simulation phase, encapsulating all lower-level objects required
    to perform a single phase (e.g., seed, initialization, simulation) of a
    given cell cluster, configuration, and simulation.

    This object principally behaves as a simple container whose:

    * Direct parent is the root-level :class:`betse.science.simrunner.SimRunner`
      object owning all objects pertaining to the current cell cluster,
      configuration, and simulation.
    * Direct children are:
      * The current cell cluster.
      * The current simulation.
      * The current simulation configuration.

    Attributes
    ----------
    kind : SimPhaseKind
        Type of this simulation phase.

    Attributes (High-level)
    ----------
    cells : betse.science.cells.Cells
        Cell cluster for this phase.
    dyna : betse.science.tissue.tishandler.TissueHandler
        Tissue handler for this phase.
    p : betse.science.parameters.Parameters
        Simulation configuration for this phase.
    sim : betse.science.sim.Simulator
        Simulation for this phase.
    cache : betse.science.phase.cache.cacheabc.SimPhaseCaches
        Simulation cache for this phase.

    Attributes (Caller)
    ----------
    callbacks : SimCallbacksABC
        Caller-defined object whose methods are periodically called during this
        phase (e.g., to notify this caller of phase progress).

    Attributes (Path)
    ----------
    export_dirname : StrOrNoneTypes
        Absolute path of the top-level directory containing all exported results
        (e.g., plots, animations, CSVs) for this simulation phase if this phase
        is either an initialization or simulation *or* ``None`` otherwise (i.e.,
        if this phase is a seed).
    '''

    # ..................{ INITIALIZORS                       }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        kind: SimPhaseKind,
        p:     'betse.science.parameters.Parameters',

        # Optional parameters.
        cells: ('betse.science.cells.Cells', NoneType) = None,
        sim:   ('betse.science.sim.Simulator', NoneType) = None,
        callbacks: SimCallbacksABCOrNoneTypes = None,
    ) -> None:
        '''
        Initialize this simulation phase.

        Parameters
        ----------
        kind : SimPhaseKind
            Current simulation phase type.
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        cells : (betse.science.cells.Cells, NoneType)
            Current cell cluster. Defaults to ``None``, in which case this
            defaults to the empty cell cluster for this configuration.
        sim : (betse.science.sim.Simulation, NoneType)
            Current simulation. Defaults to ``None``, in which case this
            defaults to an uninitialized simulation for this configuration.
        callbacks : SimCallbacksABCOrNoneTypes
            Caller-defined object whose methods are periodically called during
            this phase (e.g., to notify this caller of phase progress). Defaults
            to ``None``, in which case this defaults to a placeholder object
            whose methods all silently reduce to noops.
        '''

        # Avoid circular import dependencies.
        from betse.science.cells import Cells
        from betse.science.math.cache.cacheabc import SimPhaseCaches
        from betse.science.sim import Simulator
        from betse.science.tissue.tishandler import TissueHandler

        # Default all unpassed parameters to sane defaults.
        if callbacks is None:
            callbacks = phasecallbacks.make_default()
        if cells is None:
            cells = Cells(p=p)
        if sim is None:
            sim = Simulator(p=p)

        # Classify all passed parameters.
        self.callbacks = callbacks
        self.cells = cells
        self.kind = kind
        self.p = p
        self.sim = sim

        #FIXME: To ensure the "dyna" object is fully initialized, refactor the
        #"TissueHandler" class as follows:
        #
        #* The TissueHandler.tissueProfiles() method should be refactored into a
        #  private TissueHandler._map_tissues_to_cells() method having
        #  the following signature:
        #      @type_check
        #      def _map_tissues_to_cells(self, phase: SimPhase) -> None:
        #* The TissueHandler.__init__() method should be refactered to:
        #  * As the very first logic in that method, validate that the passed
        #    "cells" and "sim" objects have been *FULLY* initialized.
        #  * As the very last logic in that method, call the newly refactored
        #    self._map_tissues_to_cells() method.
        #* The TissueHandler._map_tissues_to_cells() method should *NOT* be
        #  called by any other class. In practice, we probably will need to
        #  break encapsulation and call the newly private
        #  TissueHandler._map_tissues_to_cells() method elsewhere. When we do
        #  so, add a FIXME comment suggesting this to be bad.
        #
        #Alternately, if the _map_tissues_to_cells() method truly *DOES* need
        #to be called elsewhere, simply make it public. *sigh*

        # Classify all remaining high-level objects for this phase.
        self.cache = SimPhaseCaches(phase=self)
        self.dyna = TissueHandler(p=p)

        #FIXME: Isolate exports produced by the "seed" phase to their own
        #directory; for simplicity, these exports currently reuse the same
        #directory as that of the "init" phase.

        # Absolute path of the top-level exports directory for this phase.
        if kind is SimPhaseKind.SEED:
            self.export_dirname = p.init_export_dirname
        elif kind is SimPhaseKind.INIT:
            self.export_dirname = p.init_export_dirname
        elif kind is SimPhaseKind.SIM:
            self.export_dirname = p.sim_export_dirname
        else:
            raise BetseSimPhaseException(
                'Simulation phase "{}" unrecognized.'.format(kind.name))

    # ..................{ EXCEPTIONS                         }..................
    @type_check
    def die_unless_kind(self, kind: SimPhaseKind) -> None:
        '''
        Raise an exception unless the kind of this simulation phase is exactly
        the passed kind of such phases (e.g., seed, initialization, simulation).
        '''

        if self.kind is not kind:
            raise BetseSimPhaseException(
                'Simulation phase "{}" not "{}".'.format(self.kind, kind))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed simulation subconfiguration classes for exporting visuals (e.g.,
plots, animations).
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from betse.lib.yaml.yamlalias import yaml_alias
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.lib.yaml.abc.yamllistabc import (
    YamlList, YamlListItemABC, YamlListItemTypedABC)
# from betse.util.io.log import logs
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.types import type_check, NumericSimpleTypes

# ....................{ SUPERCLASSES                       }....................
#FIXME: Non-ideal. Ideally, all networks subconfigurations should be refactored
#to leverage the YAML format specified by "SimConfVisualCellsYAMLMixin".
class SimConfVisualCellsABC(object, metaclass=ABCMeta):
    '''
    Abstract base class generalizing logic common to all cell cluster visual
    subconfigurations -- YAML-backed or otherwise.

    Attributes (Colorbar)
    ----------
    color_max : NumericSimpleTypes
        Maximum color value to be displayed by this visual's colorbar.
        Ignored if :attr:`is_color_autoscaled` is ``True``.
    color_min : NumericSimpleTypes
        Minimum color value to be displayed by this visual's colorbar.
        Ignored if :attr:`is_color_autoscaled` is ``True``.
    is_color_autoscaled : bool
        ``True`` if dynamically setting the minimum and maximum colorbar values
        for this visual to the minimum and maximum values flattened from
        the corresponding time series *or* ``False`` if statically setting these
        values to :attr:`color_min` and :attr:`color_max`.
    '''

    # ..................{ PROPERTIES                         }..................
    @abstractproperty
    def is_color_autoscaled(self) -> bool:
        pass

    @abstractproperty
    def color_min(self) -> NumericSimpleTypes:
        pass

    @abstractproperty
    def color_max(self) -> NumericSimpleTypes:
        pass


class SimConfVisualCellsYAMLMixin(SimConfVisualCellsABC):
    '''
    Abstract mixin generalizing implementation common to all YAML-backed cell
    cluster visual subconfiguration subclasses.

    This mixin encapsulates configuration of a single visual (either in- or
    post-simulation plot or animation) parsed from the current YAML-formatted
    simulation configuration file. For generality, this mixin provides no
    support for a YAML ``type`` key or corresponding :attr:`name` property.
    '''

    # ..................{ ALIASES ~ colorbar                 }..................
    is_color_autoscaled = yaml_alias("['colorbar']['autoscale']", bool)
    color_min = yaml_alias("['colorbar']['minimum']", NumericSimpleTypes)
    color_max = yaml_alias("['colorbar']['maximum']", NumericSimpleTypes)

# ....................{ SUBCLASSES                         }....................
#FIXME: Eliminate this subclass. For serializability, all configuration classes
#should be YAML-backed.
class SimConfVisualCellsNonYAML(SimConfVisualCellsABC):
    '''
    Cell cluster visual subconfiguration specific to the
    :mod:`betse.science.networks` package.

    Unlike all cell cluster visual subconfigurations, this class preserves
    backward compatibility by implementing superclass properties *without*
    calling the :func:`yaml_alias` function returning YAML-backed data
    descriptors. Subconfiguration changes are *not* propagated back to disk.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        is_color_autoscaled: bool,
        color_min: NumericSimpleTypes,
        color_max: NumericSimpleTypes,
    ) -> None:

        self._is_color_autoscaled = is_color_autoscaled
        self._color_min = color_min
        self._color_max = color_max

    # ..................{ PROPERTIES                         }..................
    @property
    def is_color_autoscaled(self) -> bool:
        return self._is_color_autoscaled

    @property
    def color_min(self) -> NumericSimpleTypes:
        return self._color_min

    @property
    def color_max(self) -> NumericSimpleTypes:
        return self._color_max

# ....................{ SUBCLASSES                         }....................
class SimConfVisualCellsEmbedded(SimConfVisualCellsYAMLMixin, YamlABC):
    '''
    YAML-backed cell cluster visual subconfiguration, configuring a single
    visual applicable to all cells parsed from a dictionary specifying at least
    this visual in the current YAML-formatted simulation configuration file.

    This is the *only* visual configured by this dictionary. Hence, this
    dictionary contains no distinguishing ``type`` entry and this
    subconfiguration class no corresponding :attr:`name` property.
    '''

    pass

# ....................{ SUBCLASSES : item                  }....................
class SimConfVisualCellsListItem(
    SimConfVisualCellsYAMLMixin, YamlListItemTypedABC):
    '''
    YAML-backed cell cluster visual subconfiguration list item, configuring a
    single visual applicable to all cells parsed from a list of these visuals in
    the current YAML-formatted simulation configuration file.
    '''

    # ..................{ MAKERS                             }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Duplicate the default animation listed first in our default YAML file.
        return SimConfVisualCellsListItem().load(conf={
            'type': 'voltage_membrane',
            'enabled': True,
            'colorbar': {
                'autoscale': True,
                'minimum': -70.0,
                'maximum':  10.0,
            },
        })


class SimConfVisualCellListItem(YamlListItemTypedABC):
    '''
    YAML-backed single-cell visual subconfiguration list item, configuring a
    single visual (either in- or post-simulation plot or animation) specific to
    a single cell parsed from the list of all such visuals in the current
    YAML-formatted simulation configuration file.
    '''

    # ..................{ MAKERS                             }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Duplicate the default animation listed first in our default YAML file.
        return SimConfVisualCellListItem().load(conf={
            'type': 'voltage_membrane',
            'enabled': True,
        })



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Primitive two-dimensional **convex polygon** (i.e., non-self-intersecting
polygon such that all line segments between any two points on the polygon
boundary remain strictly inside the polygon) functionality.
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.exceptions import BetseMathPolygonException
from betse.lib.numpy import nparray
from betse.science.math.geometry import geopoint
from betse.util.type.types import type_check, SequenceTypes

# ....................{ EXCEPTIONS                         }....................
def die_unless_polygon(*polygons: SequenceTypes) -> None:
    '''
    Raise an exception unless all passed sequences are two-dimensional
    polygons.

    Parameters
    ----------
    polygons : tuple[SequenceTypes]
        Tuple of all sequences to be validated.

    Raises
    ----------
    BetseMathPointException
        If any such sequence is *not* a two-dimensional polygon.

    See Also
    ----------
    :func:`is_polygon`
        Further details.
    '''

    if not is_polygon(*polygons):
        for polygon in polygons:
            if not is_polygon(polygon):
                raise BetseMathPolygonException(
                    'Sequence not a two-dimensional polygon '
                    '(i.e., length < 3 or item length != 2): {!r}'.format(
                        polygon))

# ....................{ TESTERS                            }....................
@type_check
def is_polygon(*polygons: SequenceTypes) -> bool:
    '''
    ``True`` only if all passed sequences are **two-dimensional polygons** (i.e.,
    contain at least three two-dimensional points).

    Parameters
    ----------
    polygons : tuple[SequenceTypes]
        Tuple of all sequences to be tested.

    Returns
    ----------
    bool
        ``True`` only if these sequences are all two-dimensional polygons.

    See Also
    ----------
    :func:`geopoint.is_point`
        Further details on two-dimensional points.
    '''

    return all(
        len(polygon) >= 3 and geopoint.is_point(*polygon)
        for polygon in polygons
    )

# ....................{ ORIENTERS                          }....................
#FIXME: Unit test us up.
@type_check
def orient_counterclockwise(polygon: SequenceTypes) -> SequenceTypes:
    '''
    Positively reorient the passed polygon, returning a copy of this polygon
    whose vertices are **positively oriented** (i.e., sorted in
    counter-clockwise order).

    Parameters
    ----------
    polygon : SequenceTypes
        Two-dimensional sequence of all points defining the possibly non-convex
        two-dimensional polygon to be positively oriented such that:
        * The first dimension indexes each such point (in arbitrary order).
        * The second dimension indexes each coordinate of this point such that:
          * The first item is the X coordinate of this point.
          * The second item is the Y coordinate of this point.
        Note that this function expects the type of this sequence to define an
        ``__init__()`` method accepting a passed iterable as its first and only
        positional argument. Unsurprisingly, all builtin sequences (e.g.,
        :class:`tuple`, :class:`list`) *and* :mod:`numpy` sequences (e.g.,
        :class:`numpy.ndarray`) satisfy this requirement.

    Returns
    ----------
    SequenceTypes
        Copy of this polygon positively oriented. The type of this sequence is
        the same as that of the original polygon.
    '''

    # If this sequence is *NOT* a polygon, raise an exception.
    die_unless_polygon(polygon)

    # Numpy array corresponding to this sequence. While polygon reorientation is
    # feasible in pure-Python, the Numpy-based approach is significantly more
    # efficient as the number of polygon edges increases.
    poly_verts = nparray.from_iterable(polygon)

    # Centre point of this polygon,
    poly_centre = poly_verts.mean(axis=0)

    # One-dimensional Numpy array indexing each vertex of this polygon such that
    # each element is the angle in radians between the positive X-axis and that
    # vertex, derived according to the classic mnemonic SOHCAHTOA:
    #
    #              opposite             (opposite)
    #              --------             (--------)
    #     tan(θ) = adjacent  -->  arctan(adjacent) = θ
    #
    # To sort vertices in a counter-clockwise rotation around the centre point
    # of this polygon rather than around the origin (i.e., the point (0, 0)),
    # each vertex is translated from the origin to this centre point *BEFORE*
    # obtaining this vertex's angle. Assuming standard orientation for a right
    # triangle sitting at the centre point of this polygon:
    #
    # * The opposite edge is the Y coordinate of the current vertex translated
    #   from the origin onto the centre point.
    # * The adjacent edge is the X coordinate of the current vertex translated
    #   from the origin onto the centre point.
    #
    # For safety, the np.arctan2() function intended exactly for this purpose
    # rather than the np.arctan() function intended for general-purpose
    # calculation is called. For further details, see:
    #     https://en.wikipedia.org/wiki/Atan2
    poly_angles = np.arctan2(
        poly_verts[:,1] - poly_centre[1],
        poly_verts[:,0] - poly_centre[0])

    # One-dimensional Numpy array indexing the index of each vertex of this
    # polygon, sorted in counter-clockwise order.
    poly_verts_sorted_index = poly_angles.argsort()

    # Two-dimensional Numpy array of the polygon to be returned, containing all
    # vertices sorted in this order.
    poly_verts_sorted = poly_verts[poly_verts_sorted_index]

    # Return a sequence of the same type as the passed polygon.
    return nparray.to_iterable(array=poly_verts_sorted, cls=type(polygon))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application-specific exception hierarchy.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during application startup, this module may
# import *ONLY* from modules guaranteed to exist at startup. This includes all
# standard Python and application modules but *NOT* third-party dependencies,
# which if unimportable will only be validated at some later time in startup.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from abc import ABCMeta

# ....................{ EXCEPTIONS                         }....................
#FIXME: Define an __init__() method asserting that the passed exception message
#is non-None, which Python permits by default but which is functionally useless.
class BetseException(Exception, metaclass=ABCMeta):
    '''
    Abstract base class of all application-specific exceptions.
    '''

    pass


class BetseAttributeException(BetseException):
    '''
    **Attribute** (i.e., variable or method bound to an object)-specific
    exception.
    '''

    pass


class BetseLogException(BetseException):
    '''
    Logging-specific exception.
    '''

    pass


class BetseModuleException(BetseException):
    '''
    Module-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ os                    }....................
class BetseOSException(BetseException):
    '''
    General-purpose Low-level operating system (OS) exception.
    '''

    pass


class BetseOSShellEnvException(BetseOSException):
    '''
    Shell environment-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ python                }....................
class BetsePyException(BetseException):
    '''
    General-purpose low-level Python interpreter exception.

    This exception is appropriate for use in relation to low-level issues
    concerning the active Python interpreter (e.g., inability to retrieve this
    interpreter's absolute path).
    '''

    pass


class BetsePyFrozenException(BetsePyException):
    '''
    Low-level exception pertaining to **frozen executables** (i.e., Python
    codebases converted into platform-specific executables).
    '''

    pass


class BetsePyIdentifierException(BetsePyException):
    '''
    Low-level exception pertaining to **Python identifiers** (i.e., class,
    module, or attribute name).
    '''

    pass

# ....................{ EXCEPTIONS ~ arg                  }....................
class BetseCLIException(BetseException):
    '''
    General-purpose command-line interface (CLI) exception.
    '''

    pass


class BetseCLIArgException(BetseCLIException):
    '''
    Command-line interface (CLI) argument-specific exception.
    '''

    pass


class BetseCLIArgParserException(BetseCLIArgException):
    '''
    **Argument parser** (i.e., :class:`argparse.ArgumentParser`)-specific
    exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ callable              }....................
class BetseCallableException(BetseException):
    '''
    General-purpose exception applicable to all **callables** (e.g., functions,
    lambdas, methods, properties).
    '''

    pass


class BetseCallbackException(BetseCallableException):
    '''
    Callback-specific exception.
    '''

    pass


class BetseDecoratorException(BetseCallableException):
    '''
    Decorator-specific exception.
    '''

    pass


class BetseFunctionException(BetseCallableException):
    '''
    Function-specific exception.
    '''

    pass


class BetseLambdaException(BetseCallableException):
    '''
    Lambda-specific exception.
    '''

    pass


class BetseMethodException(BetseCallableException):
    '''
    Method-specific exception.
    '''

    pass


class BetseMethodUnimplementedException(
    BetseCallableException, NotImplementedError):
    '''
    Unimplemented method-specific exception.

    This exception is typically raised from **unimplemented optional methods**
    (i.e., non-mandatory methods _not_ intended to be called) of concrete
    subclasses of abstract base classes. While the optimal solution for
    defining **unimplemented mandatory methods** (i.e., non-optional methods
    also _not_ intended to be called) is via the canonical `@abc.abstractmethod`
    decorator, there currently exists no canonical alternative for defining
    optional methods. Hence, this exception.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:

        # Avoid circular import dependencies.
        from betse.util.type.call import callers

        # Initialize our superclass with a human-readable exception message.
        super().__init__('Method {}() unimplemented.'.format(
            callers.get_caller_basename()))

# ....................{ EXCEPTIONS ~ callable : descriptor }....................
class BetseDescriptorException(BetseCallableException):
    '''
    Descriptor-specific exception.
    '''
    pass


class BetseExprAliasException(BetseDescriptorException):
    '''
    Expression alias-specific exception.
    '''
    pass


# ....................{ EXCEPTIONS ~ lib                   }....................
class BetseLibException(BetseException):
    '''
    General-purpose exception applicable to third-party dependencies.
    '''

    pass


class BetseMatplotlibException(BetseLibException):
    '''
    Matplotlib-specific exception.
    '''

    pass


class BetsePyDotException(BetseLibException):
    '''
    PyDot-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ math                  }....................
class BetseMathException(BetseException):
    '''
    General-purpose exception applicable to all low-level math algorithms.
    '''

    pass


class BetseMathLineException(BetseMathException):
    '''
    Line- and line segment-specific math exception.
    '''

    pass


class BetseMathPointException(BetseMathException):
    '''
    Point-specific math exception.
    '''

    pass


class BetseMathPolygonException(BetseMathException):
    '''
    Polygon-specific math exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ path                  }....................
class BetsePathnameException(BetseException):
    '''
    Pathname-specific exception.
    '''

    pass


class BetsePathException(BetseException):
    '''
    Path-specific exception.
    '''

    pass


class BetseArchiveException(BetsePathException):
    '''
    Archive-specific exception.
    '''

    pass


class BetseDirException(BetsePathException):
    '''
    Directory-specific exception.
    '''

    pass


class BetseGitException(BetsePathException):
    '''
    Git-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ path : file           }....................
class BetseFileException(BetsePathException):
    '''
    File-specific exception.
    '''

    pass


class BetseCommandException(BetseFileException):
    '''
    **Command** (i.e., executable binary file)-specific exception.
    '''

    pass


class BetseImageException(BetseFileException):
    '''
    Image file-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ test                  }....................
class BetseTestException(BetseException):
    '''
    General-purpose exception pertaining to this application's test suite.
    '''

    pass

# ....................{ EXCEPTIONS ~ type                  }....................
class BetseTypeException(BetseException):
    '''
    General-purpose exception applicable to types (i.e., classes).
    '''

    pass


class BetseEnumException(BetseTypeException):
    '''
    Enumeration-specific type or value exception.
    '''

    pass


class BetseNumericException(BetseTypeException):
    '''
    Exception generally applicable to both integer and float types and values.
    '''

    pass


class BetseIntException(BetseTypeException):
    '''
    Integer-specific type or value exception.
    '''

    pass


class BetseMappingException(BetseTypeException):
    '''
    Dictionary-specific type or value exception.
    '''

    pass


class BetseIterableException(BetseTypeException):
    '''
    Iterable-specific type or value exception.
    '''

    pass


class BetseSequenceException(BetseTypeException):
    '''
    Sequence-specific type or value exception.
    '''

    pass


# ....................{ EXCEPTIONS ~ type : str            }....................
class BetseStrException(BetseTypeException):
    '''
    String-specific type or value exception.
    '''

    pass


class BetseCharException(BetseStrException):
    '''
    **Character** (i.e., string of length 1)-specific type or value exception.
    '''

    pass



class BetseRegexException(BetseStrException):
    '''
    Regular exception-specific type or value exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ yaml                  }....................
class BetseYamlException(BetseException):
    '''
    Yet Another Markup Language (YAML)-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ sim                   }....................
class BetseSimException(BetseException):
    '''
    General-purpose simulation exception.
    '''

    pass


class BetseSimConfException(BetseSimException):
    '''
    Simulation configuration-specific exception.
    '''

    pass


class BetseSimPhaseException(BetseSimException):
    '''
    Simulation phase-specific exception.
    '''

    pass


class BetseSimTissueException(BetseSimException):
    '''
    Simulation tissue-specific exception.
    '''

    pass


class BetseSimVectorException(BetseSimException):
    '''
    Vector-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ sim : unstable        }....................
class BetseSimUnstableException(BetseSimException):
    '''
    Simulation-specific exception indicating the current simulation to have
    unexpectedly failed due to computational instability.
    '''

    pass


class BetseSimUnstableNaNException(BetseSimUnstableException):
    '''
    Simulation-specific exception indicating the current simulation to have
    unexpectedly failed due to a **NaN-based computational instability** (i.e.,
    an operation producing a Numpy array containing at least one NaN value,
    typically due to division by zero or infinity).
    '''

    pass

# ....................{ EXCEPTIONS ~ science : visual      }....................
class BetseSimVisualException(BetseSimException):
    '''
    Simulation visualization-specific exception, applicable to both plots and
    animations.
    '''

    pass


class BetseSimVisualLayerException(BetseSimVisualException):
    '''
    Simulation visualization layer-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ science : pipe        }....................
class BetseSimPipeException(BetseSimException):
    '''
    Simulation pipeline-specific exception.
    '''

    pass


class BetseSimPipeRunnerUnsatisfiedException(BetseSimPipeException):
    '''
    Simulation pipeline-specific exception raised on attempting to run a runner
    with unsatisfied requirements (e.g., a post-simulation animation requiring
    extracellular spaces to be enabled by the current simulation configuration).

    Attributes
    ----------
    result : str
        Human-readable string justifying this failure. For generality, this
        string is neither capitalized *nor* punctuated.
    reason : str
        Human-readable string justifying this failure. For generality, this
        string is neither capitalized *nor* punctuated.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, result: str, reason: str) -> None:
        '''
        Initialize this exception.

        Parameters
        ----------
        result : str
            Human-readable string describing this failure. For generality, this
            string is expected to be neither capitalized *nor* punctuated.
        reason : str
            Human-readable string justifying this failure. For generality, this
            string is expected to be neither capitalized *nor* punctuated.
        '''

        # Capitalize all passed parameters.
        self.result = result
        self.reason = reason

        # Initialize our superclass, concatenating these strings into a single
        # human-readable exception message.
        super().__init__('{}: {}.'.format(self.result, self.reason))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level shell-specific directory facilities.
'''

# ....................{ IMPORTS                            }....................
import os
from betse.util.io.log import logs
from betse.util.type.types import type_check
from contextlib import contextmanager
from types import GeneratorType

# ....................{ GETTERS                            }....................
def get_cwd_dirname() -> str:
    '''
    **Current working dirname** (i.e., absolute path of the current working
    directory (CWD)) of the active Python process.

    Unless subsequently changed, this is the absolute path of the directory from
    which this application was initially run.
    '''

    return os.getcwd()

# ....................{ SETTERS                            }....................
@type_check
def set_cwd(dirname: str) -> None:
    '''
    Set the **current working directory** (CWD) of the active Python process to
    the passed directory.

    This function permanently changes the CWD for the remainder of this process.
    For a robust alternative changing the CWD for a single code block, consider
    using the :func:`setting_cwd` context manager instead.

    Parameters
    -----------
    dirname : str
        Absolute or relative path of the directory to change to.
    '''

    # Log this change.
    logs.log_debug('Changing current working directory to: %s', dirname)

    # Change to this directory.
    os.chdir(dirname)

# ....................{ CONTEXTS                           }....................
@contextmanager
@type_check
def setting_cwd(dirname: str) -> GeneratorType:
    '''
    Context manager setting the **current working directory** (CWD) of the
    active Python process to the passed directory for the duration of this
    context.

    This context manager guaranteeably reverts the CWD to the prior CWD even
    when fatal exceptions are raised (e.g., due to this directory not existing).

    Parameters
    -----------
    dirname : str
        Absolute or relative path of the directory to change to.

    Returns
    -----------
    contextlib._GeneratorContextManager
        Context manager changing the CWD as described above.

    Yields
    -----------
    None
        Since this context manager yields no values, the caller's ``with``
        statement must be suffixed by *no* ``as`` clause.

    See Also
    -----------
    https://stackoverflow.com/a/24176022/2809027
        StackOverflow answer strongly inspiring this implementation.

    Examples
    -----------
    >>> from betse.util.paths import dirs
    >>> print('CWD: ' + dirs.get_cwd_dirname())
    CWD: /home/azrael
    >>> with dirs.current('/home/uriel/urial/nuriel/uryan/jeremiel'):
    ...     print('CWD: ' + dirs.get_cwd_dirname())
    ...     raise ValueError(
    ...         'But unknown, abstracted, brooding secret the dark power hid')
    CWD: /home/uriel/urial/nuriel/uryan/jeremiel
    ValueError: But unknown, abstracted, brooding secret the dark power hid.
    >>> print('CWD: ' + dirs.get_cwd_dirname())
    CWD: /home/azrael
    '''

    # Absolute path of the current CWD.
    dirname_prior = get_cwd_dirname()

    # Temporarily change to the passed directory. Since Python performs this
    # change only if this call raises no exceptions, this call need *NOT* be
    # embedded in the "try" block below.
    set_cwd(dirname)

    # Yield control to the body of the caller's "with" block.
    try:
        yield
    # Revert to the prior CWD even if that block raised an exception.
    finally:
        os.chdir(dirname_prior)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Microsoft Windows-specific facilities.

Caveats
----------
Operating system-specific logic is poor form and should be leveraged only where
necessary.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseOSException
from betse.util.type.types import type_check, NoneType

# ....................{ TYPES                              }....................
WindowsErrorType = NoneType
'''
Windows-specific :class:`WindowsError` class if the current platform is Windows
_or_ the :class:`NoneType` class otherwise.

Since all exceptions subclass the root :class:`Exception` superclass _and_
since the :class:`NoneType` class does not do so, this class is only an
exception class under Windows. In particular, attempting to either catch this
class as an exception _or_ to test whether a caught exception is an instance of
this class is guaranteed to both be safe and behave as expected regardless of
the current platform. In short: don't worry, be emoji.
'''

# For use in type checking, this class is safely defined at the top-level in a
# platform-agnostic manner.
try:
    WindowsErrorType = WindowsError
except:
    pass

# ....................{ CONSTANTS ~ error codes            }....................
# For conformance, the names of all error code constants defined below are
# exactly as specified by Microsoft itself. Sadly, Python fails to provide these
# magic numbers for us.

_ERROR_INVALID_NAME = 123
'''
Microsoft Windows-specific error code indicating an invalid pathname.

See Also
----------
https://msdn.microsoft.com/en-us/library/windows/desktop/ms681382%28v=vs.85%29.aspx
    Official listing of all such codes.
'''

# ....................{ EXCEPTIONS                         }....................
def die_unless_windows() -> None:
    '''
    Raise an exception unless the current platform is Microsoft Windows.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses

    # If the current platform is *NOT* Windows, raise an exception.
    if not oses.is_windows():
        raise BetseOSException(
            'Current platform {} not Windows.'.format(oses.get_name()))

# ....................{ TESTERS                            }....................
@type_check
def is_exception_pathname_invalid(exception: WindowsErrorType) -> bool:
    '''
    `True` only if the passed Windows-specific exception was raised from an
    erroneous attempt to read or write an invalid pathname.
    '''

    # Raise an exception unless the current platform is Windows.
    die_unless_windows()

    # The above type checking guarantees this exception to be an instance
    # of the Windows-specific "WindowsError" class. Test the "winerror"
    # attribute exclusive to such intsances.
    return exception.winerror == _ERROR_INVALID_NAME
